import argparse, os, textwrap
from pathlib import Path

from xklb.utils import (
    arg_utils,
    argparse_utils,
    consts,
    db_utils,
    file_utils,
    iterables,
    nums,
    objects,
    processes,
    sql_utils,
)
from xklb.utils.consts import DEFAULT_FILE_ROWS_READ_LIMIT, DBType
from xklb.utils.log_utils import log


def get_caller_name():
    import inspect

    frame = inspect.currentframe()
    while frame:
        frame = frame.f_back

        if frame is None:
            return "No parse_args()"
        elif frame.f_code.co_name == "parse_args":
            frame = frame.f_back
            return frame.f_code.co_name  # type: ignore


def args_post(args, parser, create_db=False):
    args.defaults = {k: v for k, v in args.__dict__.items() if parser.get_default(k) == v}
    settings = {
        k: v
        for k, v in args.__dict__.items()
        if k not in ["database", "verbose", "defaults"] + list(args.defaults.keys())
    }
    args.extractor_config = {
        k: v for k, v in settings.items() if k not in ["db", "paths", "actions", "backfill_pages"]
    } | (getattr(args, "extractor_config", None) or {})

    log_args = objects.dict_filter_bool(settings)
    if log_args:
        max_v = 140
        log.info({k: v if len(str(v)) < max_v else textwrap.shorten(str(v), max_v) for k, v in log_args.items()})

    args.action = get_caller_name()

    if create_db:
        Path(args.database).touch()
        args.db = db_utils.connect(args)
    elif getattr(args, "database", False):
        args.db = db_utils.connect(args)

    if getattr(args, "timeout", False):
        processes.timeout(args.timeout)


def printing(parser):
    printing = parser.add_argument_group("Printing")
    printing.add_argument("--to-json", action="store_true", help="Write JSONL to stdout")
    printing.add_argument("--cols", "--columns", nargs="+", help="Include specific column(s) when printing")
    printing.add_argument(
        "--print",
        "-p",
        default="",
        const="p",
        nargs="?",
        help="""Print instead of play

Printing modes
-p    # print as a table
-p a  # print an aggregate report
-p b  # print a big-dirs report (see library bigdirs -h for more info)
-p f  # print fields (defaults to path; useful for piping to utilities like xargs or GNU Parallel)
-p d  # mark deleted
-p w  # mark watched

When a printing mode is explicitly set then all rows will be fetched unless --limit is explicitly set

Some printing modes can be combined
-p df  # print fields for piping into another program and mark as deleted
-p bf  # print fields from big-dirs report""",
    )


def debug(parent_parser):
    parser = parent_parser.add_argument_group("Global options")
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="""Control the level of logging verbosity
-v     # info
-vv    # debug
-vvv   # debug, with SQL query printing
-vvvv  # debug, with external libraries logging""",
    )
    parser.add_argument("--timeout", "-T", metavar="TIME", help="Quit after x minutes")
    parser.add_argument("--threads", type=int, help="Load x files in parallel")
    parser.add_argument("--same-file-threads", type=int, help="Read the same file x times in parallel")
    parser.add_argument(
        "--ext", "-e", default=[], action=argparse_utils.ArgparseList, help="Include only specific file extensions"
    )
    printing(parent_parser)


def capability_soft_delete(parent_parser):
    parser = parent_parser.add_argument_group("Modify Metadata")
    parser.add_argument(
        "--mark-deleted", "--soft-delete", action="store_true", help="Mark matching rows as deleted (soft-delete)"
    )
    parser.add_argument("--mark-watched", action="store_true", help="Mark matching rows as watched")
    parser.add_argument("--delete-rows", action="store_true", help="Delete matching rows from database table")


def capability_delete(parent_parser):
    parser = parent_parser.add_argument_group("Delete Files")
    parser.add_argument(
        "--override-trash", "--override-rm", "--trash-cmd", default="trash", help="Custom trash command"
    )
    parser.add_argument(
        "--delete-files",
        "--trash",
        "--rm",
        action="store_true",
        help="Delete matching files from filesystem",
    )


def database(parent_parser):
    parser = parent_parser.add_argument_group("Database")
    capability_soft_delete(parent_parser)
    capability_delete(parent_parser)
    parser.add_argument("--db", "-db", help="Positional argument override")
    parser.add_argument("database")


def paths_or_stdin(parent_parser, destination=False):
    parser = parent_parser.add_argument_group("Paths")
    parser.add_argument(
        "--from-file",
        "--from-text",
        "--file",
        action="store_true",
        help="Read paths from line-delimited file(s)",
    )
    parser.add_argument("--from-json", "--json", action="store_true", help="Read JSON or JSONL from stdin")
    if destination:
        parser.add_argument("paths", nargs="+", action=argparse_utils.ArgparseArgsOrStdin)
    else:
        parser.add_argument(
            "paths", nargs="*", default=argparse_utils.STDIN_DASH, action=argparse_utils.ArgparseArgsOrStdin
        )


def sql_fs(parent_parser):
    parse_fs = parent_parser.add_argument_group("FileSystemDB SQL")
    parse_fs.add_argument(
        "--limit",
        "--queue",
        "-n",
        "-l",
        "-L",
        help="""Set play queue size
-L inf  # no limit
-L 10   # 10 files
(default: 120 media, 480 images, 7 links, 7200 downloads)""",
    )
    parse_fs.add_argument(
        "--offset",
        help="""Skip files which would have been in the queue
--offset 10  # skip 10 files""",
    )
    parse_fs.add_argument(
        "--sort",
        "-u",
        nargs="+",
        default=[],
        help="""Choose media play order
--sort duration   # play shortest media first
-u duration desc  # play longest media first

You can use multiple SQL ORDER BY expressions
-u 'subtitle_count > 0 desc' # play media that has at least one subtitle first

Prioritize large-sized media
--sort 'ntile(10000) over (order by size/duration) desc'
-u 'ntile(100) over (order by size) desc'

Sort by count of media with the same-X column (default DESC: most common to least common value)
-u same-duration
-u same-title
-u same-size
-u same-width, same-height ASC, same-fps
-u same-time_uploaded same-view_count same-upvote_ratio""",
    )
    parse_fs.add_argument(
        "--random",
        "-r",
        action="store_true",
        help="Sort by random and use performance hacks to make SQLite faster for large databases",
    )

    parse_fs.add_argument("--playlists", nargs="+", action="extend", default=[], help="Include media by playlist URLs")

    parse_fs.add_argument(
        "--where",
        "-w",
        nargs="+",
        action="extend",
        default=[],
        help="""Constrain media by arbitrary SQL expressions
--where audio_count = 2  # media which have two audio tracks
-w "language = 'eng'"    # media which have an English language tag (this could be audio _or_ subtitle)
-w subtitle_count=0      # media that doesn't have subtitles""",
    )
    parse_fs.add_argument("--fts", action=argparse.BooleanOptionalAction, default=True, help="Full Text Search mode")
    parse_fs.add_argument(
        "--include",
        "--search",
        "-s",
        nargs="+",
        action="extend",
        default=[],
        help="""Include files via search
-s happy  # path, title, or tags must match "happy"

In --fts mode (default):

Use fts syntax to search specific columns:
-s 'path : mad max'
-s "path : 'mad max'" # add "quotes" to be more strict

In --no-fts mode:

Use --where to search specific columns:
--no-fts -w 'path like "%%happy%%"'

Double spaces are equal to one space:
--no-fts -s "  ost"        # will match OST and not ghost
--no-fts -s toy story      # will match '/folder/toy/something/story.mp3'
--no-fts -s "toy  story"   # will match more strictly '/folder/toy story.mp3'""",
    )
    parse_fs.add_argument(
        "--exclude",
        "-E",
        nargs="+",
        action="extend",
        default=[],
        help="""Exclude files via search
-E sad  # path, title, or tags must not match "sad"
-w 'path not like "%%sad%%"'""",
    )
    parse_fs.add_argument(
        "--exact",
        action="store_true",
        help="""Not useful except when searching paths and excluding subpaths
-s https://files/2024/ --exact  # when you want to match the folder but not its contents""",
    )
    parse_fs.add_argument(
        "--flexible-search",
        "--or",
        "--flex",
        action="store_true",
        help="""Allow results which match only one search term
-s one two --or  # results will include /one/file.mkv and /two/file.mka
-s one two       # results will only include /one/two.mkv""",
    )

    parse_fs.add_argument("--online-media-only", "--online", action="store_true", help="Exclude local media")
    parse_fs.add_argument("--local-media-only", "--local", action="store_true", help="Exclude online media")

    parse_fs.add_argument(
        "--sizes",
        "-S",
        action="append",
        help="""Constrain media to file sizes (uses the same syntax as fd-find)
-S 6           # 6 MB exactly (not likely)
-S-6           # less than 6 MB
-S+6           # more than 6 MB
-S 6%%10       # 6 MB ±10 percent (between 5 and 7 MB)
-S+5GB -S-7GB  # between 5 and 7 GB""",
    )
    parse_fs.add_argument(
        "--bitrates",
        "-b",
        action="append",
        help="""Constrain media to bitrates
-b 6           # 6 Mbps exactly (not likely)
-b-6           # less than 6 Mbps
-b+6           # more than 6 Mbps
-b 6%%10       # 6 Mbps ±10 percent (between 5 and 7 Mbps)
-b+50KB -b-700KB  # between 50 and 700 kbit/s""",
    )

    parse_fs.add_argument(
        "--created-within",
        help="""Constrain media by time_created (newer than)
--created-within '3 days'""",
    )
    parse_fs.add_argument(
        "--created-before",
        help="""Constrain media by time_created (older than)
--created-before '3 years'""",
    )
    parse_fs.add_argument(
        "--changed-within",
        "--modified-within",
        help="""Constrain media by time_modified (newer than)
--changed-within '3 days'""",
    )
    parse_fs.add_argument(
        "--changed-before",
        "--modified-before",
        help="""Constrain media by time_modified (older than)
--changed-before '3 years'""",
    )
    parse_fs.add_argument(
        "--deleted-within",
        help="""Constrain media by time_deleted (newer than)
--deleted-within '3 days'""",
    )
    parse_fs.add_argument(
        "--deleted-before",
        help="""Constrain media by time_deleted (older than)
--deleted-before '3 years'""",
    )
    parse_fs.add_argument(
        "--downloaded-within",
        help="""Constrain media by time_downloaded (newer than)
--downloaded-within '3 days'""",
    )
    parse_fs.add_argument(
        "--downloaded-before",
        help="""Constrain media by time_downloaded (older than)
--downloaded-before '3 years'""",
    )

    parse_media = parent_parser.add_argument_group("MediaDB SQL")
    parse_media.add_argument(
        "--portrait",
        action="store_true",
        help="""Constrain media to portrait orientation video
-w 'width<height' # equivalent""",
    )
    parse_media.add_argument("--no-video", "-vn", action="store_true", help="Exclude media which have videos streams")
    parse_media.add_argument("--no-audio", "-an", action="store_true", help="Exclude media which have audio streams")
    parse_media.add_argument(
        "--no-subtitles",
        "--no-subtitle",
        "--no-subs",
        "--nosubs",
        "-sn",
        action="store_true",
        help="Exclude media which have subtitle streams",
    )
    parse_media.add_argument(
        "--subtitles", "--subtitle", "-sy", action="store_true", help="Include only media which have subtitle streams"
    )

    parse_media.add_argument(
        "--played-within",
        help="""Constrain media by time_last_played (newer than)
--played-within '3 days'""",
    )
    parse_media.add_argument(
        "--played-before",
        help="""Constrain media by time_last_played (older than)
--played-before '3 years'""",
    )

    parse_media.add_argument(
        "--partial",
        "-P",
        "--previous",
        "--recent",
        default=False,
        const="n",
        nargs="?",
        help="""Play recent partially-watched videos
--partial       # play newest first
--partial old   # play oldest first
-P o            # equivalent

-P p            # sort by percent remaining
-P t            # sort by time remaining
-P s            # skip partially watched (only show unseen)

The default time used is "last-viewed" (ie. the most recent time you closed the video)
If you want to use the "first-viewed" time (ie. the very first time you opened the video)
-P f            # use watch_later file creation time instead of modified time

You can combine most of these options, though some will override others
-P fo           # using the time you first played, play the oldest videos first
-P pt           # weighted remaining (percent * time remaining)

Print media you have partially viewed with mpv
--partial -p
-P -p          # equivalent
--partial -pa  # print an aggregate report of partially watched files""",
    )

    parse_media.add_argument(
        "--duration",
        "-d",
        action="append",
        help="""Constrain media to duration
-d 6       # 6 mins exactly
-d-6       # less than 6 mins
-d+6       # more than 6 mins
-d 6%%10    # 6 mins ±10 percent (between 5 and 7 mins)
-d+5 -d-7  # between 5 and 7 mins""",
    )
    parse_media.add_argument(
        "--duration-from-size",
        action="append",
        help="""Constrain media to duration of videos which match any size constraints
--duration-from-size +3300MB -u 'duration desc, size desc'""",
    )


def sql_fs_post(args, table_prefix="m.") -> None:
    if args.to_json:
        args.print = "p"

    args.include += getattr(args, "search", [])
    if len(args.include) == 1:
        if args.include == ["."]:
            args.include = [str(Path().cwd().resolve())]
        elif os.sep in args.include[0]:
            args.include[0] = file_utils.resolve_absolute_path(args.include[0])

    arg_utils.parse_args_limit(args)

    pl_columns = db_utils.columns(args, "playlists")
    args.playlists_sort, args.playlists_select = arg_utils.parse_args_sort(args, pl_columns)
    m_columns = db_utils.columns(args, "media")
    args.sort, args.select = arg_utils.parse_args_sort(args, m_columns)

    if args.sizes:
        args.sizes = sql_utils.parse_human_to_sql(nums.human_to_bytes, "size", args.sizes)
    if args.bitrates:
        args.bitrates = sql_utils.parse_human_to_sql(nums.human_to_bits, "size*8/duration", args.bitrates)

    if args.cols:
        args.cols = list(iterables.flatten([s.split(",") for s in args.cols]))

    args.filter_sql = []
    args.aggregate_filter_sql = []
    args.filter_bindings = {}

    aggregate_filter_columns = ["time_first_played", "time_last_played", "play_count", "playhead"]
    args.filter_sql.extend(" AND " + w for w in args.where if not any(a in w for a in aggregate_filter_columns))
    args.aggregate_filter_sql.extend(" AND " + w for w in args.where if any(a in w for a in aggregate_filter_columns))

    args.filter_sql.extend(f" AND path like '%.{ext}'" for ext in args.ext)

    if (
        "time_deleted" in m_columns
        and "deleted" not in (getattr(args, "sort_groups_by", None) or "")
        and "time_deleted" not in " ".join(args.where)
    ):
        args.filter_sql.append(f"AND COALESCE({table_prefix}time_deleted,0) = 0")

    if args.local_media_only:
        args.filter_sql.append('AND path not LIKE "http%"')
        if "time_downloaded" in m_columns:
            args.filter_sql.append("AND COALESCE(time_downloaded,1) != 0")

    if args.online_media_only:
        args.filter_sql.append('AND path LIKE "http%"')
        if "time_downloaded" in m_columns:
            args.filter_sql.append("AND COALESCE(time_downloaded,0) = 0")

    if args.sizes:
        args.filter_sql.append(" and size IS NOT NULL " + args.sizes)
    if args.bitrates:
        args.filter_sql.append(" and size IS NOT NULL " + args.bitrates)

    if args.created_within:
        args.filter_sql.append(
            f"and time_created >= cast(STRFTIME('%s', datetime( 'now', '-{nums.sql_human_time(args.created_within)}')) as int)",
        )
    if args.created_before:
        args.filter_sql.append(
            f"and time_created < cast(STRFTIME('%s', datetime( 'now', '-{nums.sql_human_time(args.created_before)}')) as int)",
        )
    if args.changed_within:
        args.filter_sql.append(
            f"and time_modified >= cast(STRFTIME('%s', datetime( 'now', '-{nums.sql_human_time(args.changed_within)}')) as int)",
        )
    if args.changed_before:
        args.filter_sql.append(
            f"and time_modified < cast(STRFTIME('%s', datetime( 'now', '-{nums.sql_human_time(args.changed_before)}')) as int)",
        )
    if args.deleted_within:
        args.filter_sql.append(
            f"and time_deleted >= cast(STRFTIME('%s', datetime( 'now', '-{nums.sql_human_time(args.deleted_within)}')) as int)",
        )
    if args.deleted_before:
        args.filter_sql.append(
            f"and time_deleted < cast(STRFTIME('%s', datetime( 'now', '-{nums.sql_human_time(args.deleted_before)}')) as int)",
        )
    if args.downloaded_within:
        args.filter_sql.append(
            f"and time_downloaded >= cast(STRFTIME('%s', datetime( 'now', '-{nums.sql_human_time(args.downloaded_within)}')) as int)",
        )
    if args.downloaded_before:
        args.filter_sql.append(
            f"and time_downloaded < cast(STRFTIME('%s', datetime( 'now', '-{nums.sql_human_time(args.downloaded_before)}')) as int)",
        )

    if getattr(args, "keep_dir", False) and Path(args.keep_dir).exists():
        args.keep_dir = Path(args.keep_dir).expanduser().resolve()
        args.filter_sql.append(f'and path not like "{args.keep_dir}%"')

    if args.no_video:
        args.filter_sql.append(" and video_count=0 ")
    if args.no_audio:
        args.filter_sql.append(" and audio_count=0 ")
    if args.subtitles:
        args.filter_sql.append(" and subtitle_count>0 ")
    if args.no_subtitles:
        args.filter_sql.append(" and subtitle_count=0 ")

    if args.played_within:
        args.aggregate_filter_sql.append(
            f"and time_last_played >= cast(STRFTIME('%s', datetime( 'now', '-{nums.sql_human_time(args.played_within)}')) as int) ",
        )
    if args.played_before:
        args.aggregate_filter_sql.append(
            f"and time_last_played < cast(STRFTIME('%s', datetime( 'now', '-{nums.sql_human_time(args.played_before)}')) as int) ",
        )

    if args.partial:
        args.aggregate_filter_sql.append(
            "AND COALESCE(time_first_played,0) = 0 " if "s" in args.partial else "AND time_first_played>0 "
        )

    if args.duration:
        args.duration = sql_utils.parse_human_to_sql(nums.human_to_seconds, "duration", args.duration)
        args.filter_sql.append(" and duration IS NOT NULL " + args.duration)

    if args.duration_from_size:
        args.duration_from_size = sql_utils.parse_human_to_sql(nums.human_to_bytes, "size", args.duration_from_size)
        args.filter_sql.append(
            " and size IS NOT NULL and duration in (select distinct duration from media where 1=1 "
            + args.duration_from_size
            + ")",
        )


def playback(parent_parser):
    parser = parent_parser.add_argument_group("Playback")
    parser.add_argument(
        "--crop",
        "--zoom",
        "--stretch",
        "--fit",
        "--fill",
        action="store_true",
        help="Crop video to fill window (useful with multiple-playback)",
    )
    parser.add_argument(
        "--loop",
        action="store_true",
        help="Loop media after reaching end (useful for curation activities like multiple-playback)",
    )
    parser.add_argument("--pause", action="store_true", help="Start media paused")
    parser.add_argument(
        "--start",
        "-vs",
        help="""Start media at a specific time
--start 35%%  # start media playback at the time of 35%% of duration; wadsworth constant""",
    )
    parser.add_argument(
        "--end",
        "-ve",
        help="""Stop media at a specific time
--end 60%%  # stop media playback at the time of 60%% of duration; avogadro constant""",
    )
    parser.add_argument("--volume", type=float, help="Set volume level before playing")

    parser.add_argument("--mpv-socket", help="Use a custom mpv socket location")
    parser.add_argument(
        "--auto-seek",
        action="store_true",
        help="""Seek playback automatically
Experimental; does not work with --multiple-playback
--auto-seek --player='mpv --pause=yes --loop=yes --start=35%%'

DEPRECIATED: do instead:
function repeatdelay
    while $argv[2..-1]
        and sleep $argv[1]
    end
end
repeatdelay 1.1 xdotool key o
where 'o' is a key that seeks the amount you want in mpv""",
    )

    parser.add_argument(
        "--override-player",
        "--player",
        help='''Use a specific player
--player "vlc --vlc-opts"''',
    )
    parser.add_argument(
        "--ignore-errors",
        "--ignoreerrors",
        "-i",
        action="store_true",
        help="Continue to the next track after a playback error (eg. YouTube video deleted)",
    )

    parser.add_argument(
        "--prefetch", type=int, default=3, help="Prepare for playback by reading some file metadata before it is needed"
    )
    parser.add_argument(
        "--prefix", default="", help="Add a prefix for file paths; eg. SSHFS mount makes paths different from normal"
    )

    parser.add_argument("--folders", "--folder", action="store_true", help="Experimental escape hatch to open folder")
    parser.add_argument(
        "--folder-glob",
        "--folderglob",
        type=int,
        default=False,
        const=10,
        nargs="?",
        help="Experimental escape hatch to open a folder glob limited to x number of files",
    )


def playback_post(args):
    from shlex import split

    if args.override_player:
        args.override_player = split(args.override_player)


def post_actions(parent_parser):
    parser = parent_parser.add_argument_group("Post-Playback Actions")
    parser.add_argument(
        "--exit-code-confirm",
        action="store_true",
        help="Use exit code bifurcation (exit 0=yes vs exit 1=no) instead of asking confirmation for a post-action in the CLI or GUI",
    )
    parser.add_argument("--gui", action="store_true", help="Ask post-action confirmation in a GUI")
    parser.add_argument("--keep-dir", "--keepdir", help='ask_move: move "kept" files to this special folder')
    parser.add_argument(
        "--post-action",
        "--action",
        "-k",
        default="keep",
        help="""Post-actions -- choose what to do after playing
--post-action keep    # do nothing after playing (default)
-k delete             # delete file after playing
-k softdelete         # mark deleted after playing

-k ask_keep           # ask whether to keep after playing
-k ask_delete         # ask whether to delete after playing

-k move               # move to "keep" dir after playing
-k ask_move           # ask whether to move to "keep" folder
The default location of the keep folder is ./keep/ (relative to the played media file)
You can change this by explicitly setting an *absolute* `keep-dir` path:
-k ask_move --keep-dir /home/my/music/keep/

-k ask_move_or_delete # ask after each whether to move to "keep" folder or delete

You can also bind keys in mpv to different exit codes. For example in input.conf:
    ; quit 5

And if you run something like:
    --cmd5 ~/bin/process_audio.py
    --cmd5 echo  # this will effectively do nothing except skip the normal post-actions via mpv shortcut

When semicolon is pressed in mpv (it will exit with error code 5) then the applicable player-exit-code command
will start with the media file as the first argument; in this case `~/bin/process_audio.py $path`.
The command will be daemonized if library exits before it completes.

To prevent confusion, normal post-actions will be skipped if the exit-code is greater than 4.
Exit-codes 0, 1, 2, 3, and 4: the external post-action will run after normal post-actions. Be careful of conflicting player-exit-code command and post-action behavior when using these!""",
    )


def post_actions_post(args):
    if args.post_action:
        args.post_action = args.post_action.replace("-", "_")


def multiple_playback(parent_parser):
    parser = parent_parser.add_argument_group("Multiple Playback")
    parser.add_argument(
        "--multiple-playback",
        "-m",
        default=False,
        nargs="?",
        const=consts.DEFAULT_MULTIPLE_PLAYBACK,
        type=int,
        help="""Play multiple files at the same time
--multiple-playback    # one per display; or two if only one display detected
--multiple-playback 4  # play four media at once, divide by available screens
-m 4 --screen-name eDP # play four media at once on specific screen
-m 4 --loop --crop     # play four cropped videos on a loop
-m 4 --hstack          # use hstack style

When using `--multiple-playback` it may be helpful to set simple window focus rules to prevent keys from accidentally being entered in the wrong mpv window (as new windows are created and capture the cursor focus).
You can set and restore your previous mouse focus setting by wrapping the command like this:

    focus-under-mouse
    library watch ... --multiple-playback 4
    focus-follows-mouse

For example in KDE:

    function focus-under-mouse
        kwriteconfig5 --file kwinrc --group Windows --key FocusPolicy FocusUnderMouse
        qdbus-qt5 org.kde.KWin /KWin reconfigure
    end

    function focus-follows-mouse
        kwriteconfig5 --file kwinrc --group Windows --key FocusPolicy FocusFollowsMouse
        kwriteconfig5 --file kwinrc --group Windows --key NextFocusPrefersMouse true
        qdbus-qt5 org.kde.KWin /KWin reconfigure
    end""",
    )
    parser.add_argument("--screen-name", help="Playback on a specific display")
    parser.add_argument("--hstack", action="store_true", help="Force videos to stack horizontally")
    parser.add_argument("--vstack", action="store_true", help="Force videos to stack vertically")
    parser.add_argument(
        "--fullscreen", "--fs", action=argparse.BooleanOptionalAction, default=True, help="Open videos in fullscreen"
    )


def multiple_playback_post(args):
    if args.multiple_playback > 1:
        args.gui = True


def extractor(parent_parser):
    parser = parent_parser.add_argument_group("Extractor")
    parser.add_argument("--no-sanitize", action="store_true", help="Don't sanitize some common URL parameters")
    parser.add_argument(
        "--insert-only", "--no-extract", "--skip-extract", action="store_true", help="Insert paths into media table"
    )  # TODO: move to its own subcommand
    parser.add_argument(
        "--insert-only-playlists",
        "--insert-only-playlist",
        "--no-extract-playlists",
        "--skip-extract-playlists",
        action="store_true",
        help="Insert paths into playlists table",
    )
    parser.add_argument("--extra", action="store_true", help="Get full metadata (takes a lot longer)")


def extractor_post(args):
    from xklb.utils.path_utils import sanitize_url

    if hasattr(args, "paths"):
        args.paths = list({s.strip() for s in args.paths})
        if not args.no_sanitize:
            args.paths = [sanitize_url(args, p) for p in args.paths]
        args.paths = iterables.conform(args.paths)


def group_folders(parent_parser):
    parser = parent_parser.add_argument_group("Group Folders")
    parser.add_argument(
        "--big-dirs",
        "--bigdirs",
        "-B",
        action="count",
        default=0,
        help="""Group media by folders
Recommended to use with -L inf and --duration or --depth filters; see `lb big-dirs -h` for more info""",
    )

    parser.add_argument(
        "--sibling", "--episode", "--episodic", action="store_true", help="Shorthand for --folder-counts '>1'"
    )
    parser.add_argument("--solo", action="store_true", help="Shorthand for --folder-counts=1")

    parser.add_argument(
        "--sort-groups-by",
        "--sort-groups",
        "--sort-by",
        nargs="+",
        help="""

--sort-groups-by 'mcda median_size,-deleted'  # sort by auto-MCDA""",
    )
    parser.add_argument("--depth", "-D", type=int, help="Folder depth of files")
    parser.add_argument("--parents", action="store_true", help="Include recursive sub-files in folder statistics")

    parser.add_argument(
        "--folder-sizes",
        "--foldersizes",
        "-FS",
        action="append",
        help="Only include folders of specific sizes (uses the same syntax as fd-find)",
    )
    parser.add_argument(
        "--folder-counts",
        "--files-counts",
        "--file-counts",
        "--files",
        "--counts",
        "--episodes",
        "-FC",
        action="append",
        help="""Number of files per folder

-FC=-3  # fewer than 3 siblings
-FC=+3  # more than 3 siblings

-FC=3  # exactly three siblings inclusive
-FC=+3 -FC=-3  # exactly three siblings inclusive

-FC=+12 -FC=-25  # between 12 and 25 files
-FC=5%%20  # 5 siblings ±20%% (4 to 6 siblings)""",
    )
    parser.add_argument("--folders-counts", action="append", help="Number of folders per folder")


def group_folders_post(args) -> None:
    if args.solo:
        args.folder_counts = ["1"]
    if args.sibling:
        args.folder_counts = ["+2"]

    if args.folder_sizes:
        args.folder_sizes = sql_utils.parse_human_to_lambda(nums.human_to_bytes, args.folder_sizes)
    if args.folder_counts:
        args.folder_counts = sql_utils.parse_human_to_lambda(int, args.folder_counts)
    if args.folders_counts:
        args.folders_counts = sql_utils.parse_human_to_lambda(int, args.folders_counts)

    if args.sort_groups_by:
        args.sort_groups_by = arg_utils.parse_ambiguous_sort(args.sort_groups_by)
        args.sort_groups_by = ",".join(args.sort_groups_by)


def cluster(parent_parser):
    parser = parent_parser.add_argument_group("Cluster")
    parser.add_argument("--cluster-sort", "--cluster", "-C", action="store_true", help="Cluster by filename TF-IDF")
    parser.add_argument("--clusters", "--n-clusters", type=int, help="Number of KMeans clusters")
    parser.add_argument(
        "--stop-words",
        "--ignore-words",
        nargs="+",
        action="extend",
        help="""Override the default "stop-word" list to ignore specific words
--stop-words the fox jumps over the moon
--stop-words (cat stop_words.txt)""",
    )

    parser.add_argument("--print-groups", "--groups", "-g", action="store_true", help="Print groups")
    parser.add_argument("--move-groups", "-M", action="store_true", help="Move groups into subfolders")

    parser.add_argument("--near-duplicates", "--similar-only", action="store_true", help="Re-group by difflib ratio")
    parser.add_argument(
        "--unique-only", action="store_true", help="Include only 'unique' lines (not including originals or duplicates)"
    )
    parser.add_argument("--exclude-unique", "--no-unique", action="store_true", help="Exclude 'unique' lines")


def related(parser):
    parser.add_argument(
        "--related",
        "-R",
        action="count",
        default=0,
        help="""Find media related to the first result
--related  # Use fts to find similar content
-R         # equivalent
-RR        # above, plus ignores most filters""",
    )


def clobber(parent_parser):
    parser = parent_parser.add_argument_group("Replace Files")
    parser.add_argument(
        "--replace",
        "--clobber",
        action=argparse.BooleanOptionalAction,
        help="Overwrite files on path conflict (default: ask to confirm)",
    )


def simulate(parser):
    parser.add_argument("--simulate", "--dry-run", action="store_true", help="Pretend to work")


def process_ffmpeg(parent_parser):
    parser = parent_parser.add_argument_group("FFMPEG Processing")
    parser.add_argument(
        "--delete-unplayable", action="store_true", help="Delete from disk any media which does not open with ffprobe"
    )

    parser.add_argument(
        "--delete-no-video", action="store_true", help="Delete files with no video instead of extracting audio"
    )
    parser.add_argument(
        "--delete-no-audio", action="store_true", help="Delete files with no audio instead of transcoding video"
    )

    parser.add_argument("--max-height", type=int, default=960)
    parser.add_argument("--max-width", type=int, default=1440)
    parser.add_argument(
        "--max-width-buffer",
        type=float,
        default=0.2,
        help="""Don't resize videos if their width is within a certain percentage of the max-width
--max-width-buffer 0.1  # within 10%% (if --max-width=1440 then anything smaller than 1584px video will be transcoded but not resized)""",
    )
    parser.add_argument(
        "--max-height-buffer",
        type=float,
        default=0.2,
        help="""Don't resize videos if their height is within a certain percentage of the max-height
--max-height-buffer 0.1  # within 10%% (if --max-height=960 then anything shorter than 1056px video will be transcoded but not resized)""",
    )

    parser.add_argument(
        "--always-split",
        "--force-split",
        action="store_true",
        help="Split all video/audio files using silence in audio track",
    )
    parser.add_argument(
        "--split-longer-than",
        help="""Only split on silence for files longer than a specific duration
--split-longer-than 37mins""",
    )
    parser.add_argument(
        "--min-split-segment",
        default=consts.DEFAULT_MIN_SPLIT,
        help="Combine segments that are smaller than this length of time",
    )

    parser.add_argument("--audio-only", action="store_true", help="Only extract audio")
    parser.add_argument(
        "--no-preserve-video",
        action="store_true",
        help="If using --audio-only delete source files even if they have video streams",
    )

    parser.add_argument("--max-image-height", type=int, default=2400)
    parser.add_argument("--max-image-width", type=int, default=2400)


def process_ffmpeg_post(args):
    args.split_longer_than = nums.human_to_seconds(args.split_longer_than)
    args.min_split_segment = nums.human_to_seconds(args.min_split_segment)


def download(parent_parser):
    parser = parent_parser.add_argument_group("Download")
    parser.add_argument(
        "--extractor-config",
        nargs=1,
        action=argparse_utils.ArgparseDict,
        default={},
        metavar="KEY=VALUE",
        help="Add key/value pairs to override or extend default extractor/downloader configuration",
    )
    parser.add_argument("--download-archive", help="yt-dlp download archive location (--video,--audio only)")
    parser.add_argument(
        "--extract-audio-ext",
        default="opus",
        help="""Custom file extension to convert to after download
--extract-audio-ext mp3""",
    )

    parser.add_argument(
        "--ignore-errors",
        "--ignoreerrors",
        "-i",
        action="store_true",
        help="Ignore some types of download errors (do not use this blindly!)",
    )
    parser.add_argument("--safe", action="store_true", help="Download only from known domains; skip generic URLs")

    parser.add_argument(
        "--retry-delay",
        default="14 days",
        help="Must be specified in SQLITE Modifiers format: N seconds, minutes, hours, days, months, or years",
    )
    parser.add_argument(
        "--force",
        action="store_true",
        help="Fetch metadata for paths even if they are already in the media table",
    )


def download_subtitle(parent_parser):
    parser = parent_parser.add_argument_group("Subtitle Download")
    parser.add_argument("--subs", action="store_true", help="Download and embed subtitles")
    parser.add_argument("--auto-subs", "--autosubs", action="store_true", help="Prefer machine-translated subtitles")
    parser.add_argument(
        "--subtitle-languages",
        "--subtitle-language",
        "--sl",
        action=argparse_utils.ArgparseList,
        help="Download specific subtitle languages",
    )


def table_like(parent_parser):
    parser = parent_parser.add_argument_group("Table-like")
    parser.add_argument(
        "--mimetype",
        "--filetype",
        help="""Treat given files as having a specific file type
--filetype csv""",
    )
    parser.add_argument(
        "--encoding",
        help="""Treat given files as having a specific encoding
--encoding utf8""",
    )
    parser.add_argument("--table-name", "--table", "-t", help="Load a specific table by name")
    parser.add_argument("--table-index", type=int, help="Load a specific table by index")
    parser.add_argument("--start-row", "--skiprows", type=int, default=None, help="Skip reading x rows")
    parser.add_argument(
        "--end-row",
        "--nrows",
        "--limit",
        "-L",
        default=str(DEFAULT_FILE_ROWS_READ_LIMIT),
        help="Stop reading after x rows",
    )
    parser.add_argument(
        "--join-tables",
        "--concat",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="Concat all detected tables",
    )


def table_like_post(args):
    if args.end_row.lower() in ("inf", "none", "all"):
        args.end_row = None
    else:
        args.end_row = int(args.end_row)


def filter_links(parent_parser):
    parser = parent_parser.add_argument_group("Filter Links")
    parser.add_argument(
        "--path-include",
        "--include-path",
        nargs="*",
        default=[],
        help="path substrings for inclusion (all must match to include)",
    )
    parser.add_argument(
        "--text-include",
        "--include-text",
        nargs="*",
        default=[],
        help="link text substrings for inclusion (all must match to include)",
    )
    parser.add_argument(
        "--after-include",
        "--include-after",
        nargs="*",
        default=[],
        help="plain text substrings after URL for inclusion (all must match to include)",
    )
    parser.add_argument(
        "--before-include",
        "--include-before",
        nargs="*",
        default=[],
        help="plain text substrings before URL for inclusion (all must match to include)",
    )
    parser.add_argument(
        "--path-exclude",
        "--exclude-path",
        nargs="*",
        default=["javascript:", "mailto:", "tel:"],
        help="path substrings for exclusion (any must match to exclude)",
    )
    parser.add_argument(
        "--text-exclude",
        "--exclude-text",
        nargs="*",
        default=[],
        help="link text substrings for exclusion (any must match to exclude)",
    )
    parser.add_argument(
        "--after-exclude",
        "--exclude-after",
        nargs="*",
        default=[],
        help="plain text substrings after URL for exclusion (any must match to exclude)",
    )
    parser.add_argument(
        "--before-exclude",
        "--exclude-before",
        nargs="*",
        default=[],
        help="plain text substrings before URL for exclusion (any must match to exclude)",
    )

    parser.add_argument("--strict-include", action="store_true", help="All include args must resolve true")
    parser.add_argument("--strict-exclude", action="store_true", help="All exclude args must resolve true")
    parser.add_argument("--case-sensitive", action="store_true", help="Filter with case sensitivity")
    parser.add_argument(
        "--no-url-decode",
        "--skip-url-decode",
        action="store_true",
        help="Skip URL-decode for --path-include/--path-exclude",
    )
    parser.add_argument("--local-html", action="store_true", help="Treat paths as Local HTML files")


def filter_links_post(args):
    if not args.case_sensitive:
        args.before_include = [s.lower() for s in args.before_include]
        args.path_include = [s.lower() for s in args.path_include]
        args.text_include = [s.lower() for s in args.text_include]
        args.after_include = [s.lower() for s in args.after_include]
        args.before_exclude = [s.lower() for s in args.before_exclude]
        args.path_exclude = [s.lower() for s in args.path_exclude]
        args.text_exclude = [s.lower() for s in args.text_exclude]
        args.after_exclude = [s.lower() for s in args.after_exclude]

    if not args.no_url_decode:
        from xklb.utils.web import url_decode

        args.path_include = [url_decode(s) for s in args.path_include]
        args.path_exclude = [url_decode(s) for s in args.path_exclude]


def requests(parent_parser):
    parser = parent_parser.add_argument_group("Requests")
    parser.add_argument(
        "--cookies-from-browser",
        metavar="BROWSER[+KEYRING][:PROFILE][::CONTAINER]",
        help="""Load cookies from your browser
--cookies-from-browser firefox
--cookies-from-browser chrome
(uses the same syntax as yt-dlp)""",
    )
    parser.add_argument("--cookies", help="path to a Netscape formatted cookies file")
    parser.add_argument(
        "--allow-insecure",
        "--allow-untrusted",
        "--disable-tls",
        action="store_true",
        help='Allow loading data from non-TLS, non-"https" servers',
    )
    parser.add_argument("--http-max-retries", "--https-max-retries", type=int, default=8, help="Use x retries")


def selenium(parent_parser):
    parser = parent_parser.add_argument_group("Selenium")
    parser.add_argument("--selenium", "--js", action="store_true", help="Use selenium")
    parser.add_argument("--firefox", action="store_true", help="Use selenium with firefox")
    parser.add_argument("--chrome", action="store_true", help="Use selenium with chromium etc")
    parser.add_argument("--scroll", action="store_true", help="Scroll down the page; infinite scroll")
    parser.add_argument("--manual", action="store_true", help="Confirm manually in shell before exiting the browser")
    parser.add_argument(
        "--auto-pager", "--autopager", action="store_true", help="Use an auto-pager plugin to load additional pages"
    )
    parser.add_argument("--poke", action="store_true", help="Find a filled-in search box and press enter in it")


def selenium_post(args):
    if args.scroll or args.firefox or args.chrome or args.auto_pager or args.poke:
        args.selenium = True


def sample_hash_bytes(parent_parser):
    parser = parent_parser.add_argument_group("Sample Hash")
    parser.add_argument(
        "--chunk-size",
        type=int,
        help="Chunk size in bytes (default is 1%%~0.2%% dependent on file size). If set, recommended to use at least 1048576 (for performance)",
    )
    parser.add_argument(
        "--gap",
        default="0.1",
        help="Width between chunks to skip (default 10%%). Values greater than 1 are treated as number of bytes",
    )


def sample_hash_bytes_post(args):
    args.gap = nums.float_from_percent(args.gap)


def media_check(parent_parser):
    parser = parent_parser.add_argument_group("Media Check")
    parser.add_argument(
        "--chunk-size",
        type=float,
        help="Chunk size in seconds (default 0.5 second). If set, recommended to use >0.1 seconds",
        default=0.5,
    )
    parser.add_argument(
        "--gap",
        default="0.05",
        help="Width between chunks to skip (default 5%%). Values greater than 1 are treated as number of seconds",
    )
    parser.add_argument(
        "--delete-corrupt",
        "--delete-corruption",
        help="Delete media that is more corrupt or equal to this threshold. Values greater than 1 are treated as number of seconds",
    )
    parser.add_argument(
        "--full-scan-if-corrupt",
        "--full-scan-if-corruption",
        help="Full scan as second pass if initial scan result more corruption or equal to this threshold. Values greater than 1 are treated as number of seconds",
    )
    parser.add_argument("--full-scan", action="store_true", help="Decode the full media file")
    parser.add_argument("--audio-scan", action="store_true", help="Count errors in audio track only")


def media_check_post(args):
    args.gap = nums.float_from_percent(args.gap)
    if args.delete_corrupt:
        args.delete_corrupt = nums.float_from_percent(args.delete_corrupt)
    if args.full_scan_if_corrupt:
        args.full_scan_if_corrupt = nums.float_from_percent(args.full_scan_if_corrupt)


def db_profiles(parser):
    profiles = parser.add_argument_group("DB Profiles")
    profiles.add_argument(
        "--audio",
        action="append_const",
        dest="profiles",
        const=DBType.audio,
        help="Extract audio metadata",
    )
    profiles.add_argument(
        "--filesystem",
        "--fs",
        action="append_const",
        dest="profiles",
        const=DBType.filesystem,
        help="Extract filesystem metadata",
    )
    profiles.add_argument(
        "--video",
        action="append_const",
        dest="profiles",
        const=DBType.video,
        help="Extract video metadata",
    )
    profiles.add_argument(
        "--text",
        action="append_const",
        dest="profiles",
        const=DBType.text,
        help="Extract text metadata",
    )
    profiles.add_argument(
        "--image",
        action="append_const",
        dest="profiles",
        const=DBType.image,
        help="Extract image metadata",
    )
    parser.add_argument("--scan-all-files", action="store_true")


def frequency(parser):
    parser.add_argument(
        "--frequency",
        "--freqency",
        "-f",
        default="monthly",
        const="monthly",
        type=str.lower,
        nargs="?",
        help=f"One of: {', '.join(consts.frequency)} (default: %(default)s)",
    )


def frequency_post(args):
    from xklb.utils.strings import partial_startswith

    args.frequency = partial_startswith(args.frequency, consts.frequency)


def history(parser):
    parser = parser.add_argument_group("History")
    history = parser.add_mutually_exclusive_group()
    history.add_argument(
        "--completed",
        "--played",
        "--watched",
        "--listened",
        action="store_true",
        help="Exclude partially watched media",
    )
    history.add_argument(
        "--in-progress",
        "--playing",
        "--watching",
        "--listening",
        action="store_true",
        help="Exclude completely watched media",
    )



import argparse, os, shlex, subprocess, sys
from pathlib import Path

from xklb import usage
from xklb.mediafiles import process_image
from xklb.utils import arggroups, argparse_utils, nums, path_utils, processes, web
from xklb.utils.arg_utils import gen_paths, kwargs_overwrite
from xklb.utils.log_utils import log


def parse_args() -> argparse.Namespace:
    parser = argparse_utils.ArgumentParser(prog="library process-ffmpeg", usage=usage.process_ffmpeg)
    arggroups.simulate(parser)
    arggroups.process_ffmpeg(parser)
    arggroups.debug(parser)

    arggroups.paths_or_stdin(parser)
    args = parser.parse_args()
    arggroups.args_post(args, parser)

    arggroups.process_ffmpeg_post(args)

    return args


def is_animation_from_probe(probe) -> bool:
    if probe.audio_streams:
        return True
    for stream in probe.video_streams:
        frames = nums.safe_int(stream["nb_frames"])
        if frames is None:
            r = processes.cmd(
                "ffprobe",
                "-nostdin",
                "-v",
                "error",
                "-count_frames",
                "-select_streams",
                "v:0",
                "-show_entries",
                "stream=nb_read_frames",
                "-of",
                "default=nokey=1:noprint_wrappers=1",
                probe.path,
            )
            frames = nums.safe_int(r.stdout)
            if frames is None:
                raise RuntimeError

        if frames > 1:
            return True

    return False


def process_path(args, path, **kwargs):
    if kwargs:
        args = kwargs_overwrite(args, kwargs)

    output_path = Path(web.url_to_local_path(path) if str(path).startswith("http") else path)
    output_path = Path(path_utils.clean_path(bytes(output_path), max_name_len=251))

    path = Path(path)
    original_stats = path.stat()
    probe = processes.FFProbe(path)

    if not probe.streams:
        log.error("No media streams found: %s", path)
        return path

    if path.suffix.lower() in (".gif", ".png", ".apng", ".webp", ".avif", ".avifs", ".flif", ".mng"):
        is_animation = is_animation_from_probe(probe)
        if not is_animation:
            return process_image.process_path(args, path)

    def is_album_art(s):
        return s.get("disposition", {}).get("attached_pic", 0) == 1

    video_stream = next((s for s in probe.video_streams if not is_album_art(s)), None)
    audio_stream = next((s for s in probe.audio_streams), None)
    if not video_stream:
        log.warning("No video stream found: %s", path)
        if args.delete_no_video:
            path.unlink()
            return None
    if not audio_stream:
        log.warning("No audio stream found: %s", path)
        if args.delete_no_audio:
            path.unlink()
            return None

    if video_stream and not args.audio_only:
        output_suffix = ".av1.mkv"
    elif not video_stream and not audio_stream:
        output_suffix = ".mkv"
    else:
        output_suffix = ".mka"
    output_path = output_path.with_suffix(output_suffix)

    if path == output_path:
        log.error("Input and output files have the same names %s", path)
        return path

    ff_opts: list[str] = []

    if video_stream:
        ff_opts.extend(["-c:v", "libsvtav1", "-preset", "8", "-crf", "44"])

        width = int(video_stream.get("width"))
        height = int(video_stream.get("height"))

        if width > (args.max_width * (1 + args.max_width_buffer)):
            ff_opts.extend(["-vf", f"scale={args.max_width}:-2"])
        elif height > (args.max_height * (1 + args.max_height_buffer)):
            ff_opts.extend(["-vf", f"scale=-2:{args.max_height}"])

    is_split = bool(audio_stream)
    if audio_stream:
        channels = audio_stream.get("channels") or 2
        bitrate = int(audio_stream.get("bit_rate") or probe.format.get("bit_rate") or 256000)
        source_rate = int(audio_stream.get("sample_rate") or 44100)

        duration = float(audio_stream.get("duration") or probe.format.get("duration") or 0)
        is_split = args.always_split or (args.split_longer_than and duration > args.split_longer_than)

        try:
            assert bitrate > 0
            assert channels > 0
            assert source_rate > 0
        except AssertionError:
            log.exception("Broken file or audio format misdetected: %s", path)
            if args.delete_no_audio:
                path.unlink()
                return None
        else:
            if channels == 1:
                ff_opts.extend(["-ac", "1"])
            else:
                ff_opts.extend(["-ac", "2"])

            if bitrate >= 256000:
                ff_opts.extend(["-b:a", "128k"])
            else:
                ff_opts.extend(["-b:a", "64k", "-frame_duration", "40"])

            if source_rate >= 44100:
                opus_rate = 48000
            elif source_rate >= 22050:
                opus_rate = 24000
            else:
                opus_rate = 16000
            ff_opts.extend(["-c:a", "libopus", "-ar", str(opus_rate), "-filter:a", "loudnorm=i=-18:tp=-3:lra=17"])

        if is_split:
            try:
                result = subprocess.check_output(
                    [
                        "ffmpeg",
                        "-v",
                        "warning",
                        "-i",
                        path,
                        "-af",
                        "silencedetect=-55dB:d=0.3,ametadata=mode=print:file=-:key=lavfi.silence_start",
                        "-vn",
                        "-sn",
                        "-f",
                        "s16le",
                        "-y",
                        "/dev/null",
                    ]
                )
            except subprocess.CalledProcessError:
                log.exception("Splits could not be identified. Likely broken file: %s", path)
                if args.delete_unplayable:
                    path.unlink()
                    return None
                raise

            splits = result.decode().split("\n")
            splits = [line.split("=")[1] for line in splits if "lavfi.silence_start" in line]

            prev = 0.0
            final_splits = []
            for split in splits:
                split = float(split)
                if (split - prev) >= args.min_split_segment:  # type: ignore
                    final_splits.append(str(split))
                    prev = split

            if final_splits:
                output_path = path.with_suffix(".%03d" + output_suffix)
                final_splits = ",".join(final_splits)
                print(f"Splitting {path} at points: {final_splits}")
                ff_opts.extend(["-f", "segment", "-segment_times", final_splits])
            else:
                is_split = False

    if path.parent != output_path.parent:
        log.warning("Output folder will be different due to path cleaning: %s", output_path.parent)
        output_path.parent.mkdir(exist_ok=True, parents=True)

    command = [
        "ffmpeg",
        "-nostdin",
        "-hide_banner",
        "-loglevel",
        "warning",
        "-y",
        "-i",
        str(path),
        *ff_opts,
        str(output_path),
    ]

    if args.simulate:
        print(shlex.join(command))
        return path
    else:
        try:
            processes.cmd(*command)
        except subprocess.CalledProcessError:
            log.exception("Could not transcode: %s", path)
            if args.delete_unplayable:
                path.unlink()
                return None
            else:
                raise

        if is_split:
            output_path = output_path.with_name(
                output_path.name.replace(".%03d", ".000")
            )  # TODO: support / return multiple paths...

        if not Path(output_path).exists() or output_path.stat().st_size == 0:
            output_path.unlink()  # Remove transcode
            return path

        delete_original = True
        delete_transcode = False

        if video_stream and args.audio_only and not args.no_preserve_video:
            delete_original = False

        if output_path.stat().st_size > path.stat().st_size:
            delete_original = False
            delete_transcode = True

        if delete_transcode:
            output_path.unlink()
            return path
        elif delete_original:
            path.unlink()

        os.utime(output_path, (original_stats.st_atime, original_stats.st_mtime))

    return output_path


def process_ffmpeg():
    args = parse_args()

    for path in gen_paths(args):
        if not path.startswith("http"):
            path = str(Path(path).resolve())

        try:
            process_path(args, path)
        except Exception:
            print(path)
            raise


def process_audio():
    sys.argv += ["--audio-only"]
    process_ffmpeg()


if __name__ == "__main__":
    process_ffmpeg()



import argparse, os, sys

from xklb import usage
from xklb.createdb import gallery_backend, tube_backend
from xklb.mediadb import db_media
from xklb.playback import media_printer
from xklb.utils import (
    arg_utils,
    arggroups,
    argparse_utils,
    consts,
    db_utils,
    iterables,
    printing,
    processes,
    sql_utils,
    web,
)
from xklb.utils.consts import DBType
from xklb.utils.log_utils import log
from xklb.utils.sqlgroups import construct_download_query


def parse_args():
    parser = argparse_utils.ArgumentParser(
        prog="library download",
        usage=usage.download,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    arggroups.sql_fs(parser)

    arggroups.download(parser)
    arggroups.download_subtitle(parser)
    arggroups.requests(parser)

    profile = parser.add_mutually_exclusive_group()
    profile.add_argument(
        "--audio",
        action="store_const",
        dest="profile",
        const=DBType.audio,
        help="Use audio downloader",
    )
    profile.add_argument(
        "--video",
        action="store_const",
        dest="profile",
        const=DBType.video,
        help="Use video downloader",
    )
    profile.add_argument(
        "--image",
        "--photo",
        action="store_const",
        dest="profile",
        const=DBType.image,
        help="Use image downloader",
    )
    profile.add_argument(
        "--filesystem",
        "--fs",
        "--web",
        action="store_const",
        dest="profile",
        const=DBType.filesystem,
        help="Use filesystem downloader",
    )

    parser.add_argument("--prefix", default=os.getcwd())

    parser.add_argument("--same-domain", action="store_true", help="Choose a random domain to focus on")

    parser.add_argument("--small", action="store_true", help="Video: Prefer 480p-like")

    parser.add_argument("--photos", action="store_true", help="Image: Download JPG and WEBP")
    parser.add_argument("--drawings", action="store_true", help="Image: Download PNG")
    parser.add_argument("--gifs", action="store_true", help="Image: Download MP4 and GIFs")
    arggroups.debug(parser)

    arggroups.database(parser)
    arggroups.paths_or_stdin(parser)

    parser.set_defaults(paths=None, fts=False)
    args, unk = parser.parse_known_intermixed_args()
    arggroups.args_post(args, parser)

    if unk and not args.profile in (DBType.video, DBType.audio):
        parser.error(f"unrecognized arguments: {' '.join(unk)}")
    args.unk = unk

    if not args.profile and not args.print:
        log.error("Download profile must be specified. Use one of: --video OR --audio OR --image OR --filesystem")
        raise SystemExit(1)

    arggroups.sql_fs_post(args)
    return args


def mark_download_attempt(args, paths) -> int:
    paths = iterables.conform(paths)

    modified_row_count = 0
    if paths:
        df_chunked = iterables.chunks(paths, consts.SQLITE_PARAM_LIMIT)
        for chunk_paths in df_chunked:
            with args.db.conn:
                cursor = args.db.conn.execute(
                    f"""update media
                    set time_modified={consts.now()}
                    where path in ("""
                    + ",".join(["?"] * len(chunk_paths))
                    + ")",
                    (*chunk_paths,),
                )
                modified_row_count += cursor.rowcount

    return modified_row_count


def dl_download(args=None) -> None:
    if args:
        sys.argv = ["lb", *args]

    args = parse_args()
    m_columns = db_utils.columns(args, "media")

    if "limit" in args.defaults and "media" in args.db.table_names() and "webpath" in m_columns:
        if args.db.pop("SELECT 1 from media WHERE webpath is NULL and path in (select webpath from media) LIMIT 1"):
            with args.db.conn:
                args.db.conn.execute(
                    """
                    DELETE from media WHERE webpath is NULL
                    AND path in (
                        select webpath from media
                        WHERE error IS NULL OR error != 'Media check failed'
                    )
                    """
                )

    args.blocklist_rules = []
    if "blocklist" in args.db.table_names():
        args.blocklist_rules = [{d["key"]: d["value"]} for d in args.db["blocklist"].rows]

    if args.profile == DBType.filesystem:
        web.requests_session(args)  # prepare requests session

    media = list(arg_utils.gen_d(args))
    if not media:
        query, bindings = construct_download_query(args)
        media = list(args.db.query(query, bindings))

    if not media:
        processes.no_media_found()

    if args.print:
        media_printer.media_printer(args, media)
        return

    for m in media:
        if args.blocklist_rules and sql_utils.is_blocked_dict_like_sql(m, args.blocklist_rules):
            mark_download_attempt(args, [m["path"]])
            continue

        if args.safe:
            if (args.profile in (DBType.audio, DBType.video) and not tube_backend.is_supported(m["path"])) or (
                args.profile in (DBType.image) and not gallery_backend.is_supported(args, m["path"])
            ):
                log.info("[%s]: Skipping unsupported URL (safe_mode)", m["path"])
                mark_download_attempt(args, [m["path"]])
                continue

        # check if download already attempted recently by another process
        previous_time_attempted = m.get("time_modified") or consts.APPLICATION_START  # 0 is nullified
        if not args.force and "time_modified" in m_columns:
            d = args.db.pop_dict(
                f"""
                SELECT time_modified, time_deleted from media m
                WHERE 1=1
                AND path=?
                AND (time_modified > {str(previous_time_attempted)} OR time_deleted > 0)
                """,
                [m["path"]],
            )
            log.debug(d)
            if d:
                if d["time_deleted"]:
                    log.info(
                        "[%s]: Download marked deleted (%s ago). Skipping!",
                        m["path"],
                        printing.human_duration(consts.now() - d["time_deleted"]),
                    )
                    mark_download_attempt(args, [m["path"]])
                    continue
                elif d["time_modified"]:
                    log.info(
                        "[%s]: Download already attempted recently (%s ago). Skipping!",
                        m["path"],
                        printing.human_duration(consts.now() - d["time_modified"]),
                    )
                    continue

        try:
            log.debug(m)

            if args.profile in (DBType.audio, DBType.video):
                tube_backend.download(args, m)
            elif args.profile == DBType.image:
                gallery_backend.download(args, m)
            elif args.profile == DBType.filesystem:
                local_path = web.download_url(m["path"], output_prefix=args.prefix)
                db_media.download_add(args, m["path"], m, local_path)
            else:
                raise NotImplementedError
        except Exception:
            print("db:", args.database)
            raise



import concurrent.futures, json, random, sys, time
from urllib.parse import urlparse

from xklb import usage
from xklb.createdb import av, fs_add
from xklb.files import sample_hash
from xklb.mediadb import db_media, db_playlists
from xklb.text import extract_links
from xklb.utils import (
    arg_utils,
    arggroups,
    argparse_utils,
    consts,
    db_utils,
    file_utils,
    iterables,
    nums,
    printing,
    sql_utils,
    strings,
    web,
)
from xklb.utils.consts import DBType
from xklb.utils.log_utils import log


def parse_args(action, **kwargs):
    parser = argparse_utils.ArgumentParser(**kwargs)
    arggroups.db_profiles(parser)
    arggroups.requests(parser)
    arggroups.selenium(parser)
    arggroups.filter_links(parser)
    arggroups.extractor(parser)
    parser.set_defaults(threads=4)

    parser.add_argument("--hash", action="store_true")
    parser.add_argument(
        "--sizes",
        action="append",
        help="Only grab extended metadata for files of specific sizes (uses the same syntax as fd-find)",
    )

    arggroups.debug(parser)
    arggroups.database(parser)
    if "add" in action:
        arggroups.paths_or_stdin(parser)
    args = parser.parse_intermixed_args()
    arggroups.args_post(args, parser, create_db=action == consts.SC.web_add)

    if not args.profiles:
        args.profiles = [DBType.filesystem]

    if args.sizes:
        args.sizes = sql_utils.parse_human_to_lambda(nums.human_to_bytes, args.sizes)

    if hasattr(args, "paths"):
        args.paths = [strings.strip_enclosing_quotes(s) for s in iterables.conform(args.paths)]

    arggroups.extractor_post(args)
    arggroups.filter_links_post(args)
    arggroups.selenium_post(args)

    return args


def consolidate_media(args, path: str) -> dict:
    return {
        "playlists_id": args.playlists_id,
        "time_created": consts.APPLICATION_START,
        "time_deleted": 0,
        "path": path,
    }


def add_media(args, media):
    media = iterables.list_dict_filter_bool(media)
    args.db["media"].insert_all(media, pk="id", alter=True, replace=True)


def add_extra_metadata(args, m):
    extension = m["path"].rsplit(".", 1)[-1].lower()

    remote_path = m["path"]  # for temp file extraction
    if DBType.video in args.profiles and (extension in consts.VIDEO_EXTENSIONS or args.scan_all_files):
        m |= av.munge_av_tags(args, m["path"])
    if DBType.audio in args.profiles and (extension in consts.AUDIO_ONLY_EXTENSIONS or args.scan_all_files):
        m |= av.munge_av_tags(args, m["path"])
    if DBType.text in args.profiles and (extension in consts.TEXTRACT_EXTENSIONS or args.scan_all_files):
        with web.PartialContent(m["path"]) as temp_file_path:
            m |= fs_add.munge_book_tags_fast(temp_file_path)
    if DBType.image in args.profiles and (extension in consts.IMAGE_EXTENSIONS or args.scan_all_files):
        with web.PartialContent(m["path"], max_size=32 * 1024) as temp_file_path:
            m |= fs_add.extract_image_metadata_chunk([{"path": temp_file_path}])[0]
    m["path"] = remote_path  # restore from temp file extraction

    return m


def add_basic_metadata(args, m):
    if DBType.filesystem in args.profiles:
        m |= web.stat(m["path"])
        m["type"] = file_utils.mimetype(m["path"])
    else:
        extension = m["path"].rsplit(".", 1)[-1].lower()
        if (
            args.scan_all_files
            or (DBType.video in args.profiles and extension in consts.VIDEO_EXTENSIONS)
            or (DBType.audio in args.profiles and extension in consts.AUDIO_ONLY_EXTENSIONS)
            or (DBType.text in args.profiles and extension in consts.TEXTRACT_EXTENSIONS)
            or (DBType.image in args.profiles and extension in consts.IMAGE_EXTENSIONS)
        ):
            m |= web.stat(m["path"])

    if getattr(args, "hash", False):
        # TODO: use head_foot_stream
        m["hash"] = sample_hash.sample_hash_file(m["path"])

    return m


def spider(args, paths: set):
    original_paths = paths.copy()
    get_urls = iterables.return_unique(extract_links.get_inner_urls)

    new_media_count = 0
    known_paths = set()
    traversed_paths = set()
    while len(paths) > 0:
        new_paths = {}
        path = paths.pop()
        traversed_paths.add(path)
        log.info("Loading %s", path)

        printing.print_overwrite(
            f"Pages to scan {len(paths)} link scan: {new_media_count} new [{len(known_paths)} known]"
        )

        if path in original_paths or web.is_index(path) or web.is_html(path):
            for a_ref in get_urls(args, path):
                if a_ref is None:
                    break

                link, link_text = a_ref

                link = web.remove_apache_sorting_params(link)

                if link in (paths | traversed_paths):
                    continue
                if web.is_index(link):
                    if web.is_subpath(path, link):
                        paths.add(link)
                    continue

                if db_media.exists(args, link):
                    known_paths.add(link)
                    continue
                else:
                    new_paths[link] = link_text
        else:
            if path not in (paths | traversed_paths):
                if db_media.exists(args, path):
                    known_paths.add(path)
                else:
                    new_paths[path] = None  # add key to map; title: None

        media = [consolidate_media(args, k) | {"title": v} for k, v in new_paths.items()]
        new_media_count += len(media)

        with concurrent.futures.ThreadPoolExecutor(max_workers=args.threads) as executor:
            gen_media = (f.result() for f in [executor.submit(add_basic_metadata, args, m) for m in media])
            for i, m in enumerate(gen_media):
                media[i] = m
                printing.print_overwrite(
                    f"Pages to scan {len(paths)} link scan: {new_media_count} new [{len(known_paths)} known]; basic metadata {i + 1} of {len(media)}"
                )

        if args.sizes:
            basic_media, extra_media = [], []
            for d in media:
                if d.get("size") is None or args.sizes(d["size"]):
                    extra_media.append(d)
                else:
                    basic_media.append(d)
            if basic_media:
                add_media(args, basic_media)

        with concurrent.futures.ThreadPoolExecutor(max_workers=args.threads) as executor:
            gen_media = (f.result() for f in [executor.submit(add_extra_metadata, args, m) for m in media])
            for i, m in enumerate(gen_media):
                media[i] = m
                printing.print_overwrite(
                    f"Pages to scan {len(paths)} link scan: {new_media_count} new [{len(known_paths)} known]; extra metadata {i + 1} of {len(media)}"
                )

        if media:
            add_media(args, media)

        printing.print_overwrite(
            f"Pages to scan {len(paths)} link scan: {new_media_count} new [{len(known_paths)} known]"
        )

    return new_media_count


def add_playlist(args, path):
    info = {
        "hostname": urlparse(path).hostname,
        "extractor_key": "WebFolder",
        "extractor_config": args.extractor_config,
        "time_deleted": 0,
    }
    return db_playlists.add(args, str(path), info)


def web_add(args=None) -> None:
    if args:
        sys.argv = ["lb", *args]

    args = parse_args(consts.SC.web_add, usage=usage.web_add)
    web.requests_session(args)  # configure session

    if args.insert_only:
        media_new = set()
        media_known = set()
        for p in arg_utils.gen_paths(args):
            if db_media.exists(args, p):
                media_known.add(p)
            else:
                add_media(args, [consolidate_media(args, p)])
                media_new.add(p)
            printing.print_overwrite(f"Link import: {len(media_new)} new [{len(media_known)} known]")
    else:
        if args.selenium:
            web.load_selenium(args)
        try:
            for playlist_path in arg_utils.gen_paths(args):
                args.playlists_id = add_playlist(args, playlist_path)
                spider(args, {playlist_path})

        finally:
            if args.selenium:
                web.quit_selenium(args)

    if not args.db["media"].detect_fts():
        db_utils.optimize(args)


def web_update(args=None) -> None:
    if args:
        sys.argv = ["lb", *args]

    args = parse_args(consts.SC.web_add, usage=usage.web_update)
    web.requests_session(args)  # configure session

    web_playlists = db_playlists.get_all(
        args,
        sql_filters="extractor_key = 'WebFolder'",
        order_by="""length(path)-length(REPLACE(path, '/', '')) desc
        , random()
        """,
    )

    selenium_needed = any([json.loads(d.get("extractor_config") or "{}").get("selenium") for d in web_playlists])
    if selenium_needed:
        web.load_selenium(args)

    try:
        playlist_count = 0
        for playlist in web_playlists:
            extractor_config = json.loads(playlist.get("extractor_config") or "{}")
            args_env = arg_utils.override_config(args, extractor_config)

            # TODO: use directory Last-Modified header to skip file trees which don't need to be updated
            new_media = spider(args_env, {playlist["path"]})

            if new_media > 0:
                db_playlists.decrease_update_delay(args, playlist["path"])
            else:
                db_playlists.increase_update_delay(args, playlist["path"])

            if playlist_count > 3:
                time.sleep(random.uniform(0.05, 2))
            playlist_count += 1

    finally:
        if selenium_needed:
            web.quit_selenium(args)



#!/usr/bin/env python3

# Copyright (c) 2016, Antonio SJ Musumeci <trapexit@spawn.link>
# Copyright (c) 2024, Jacob Chapman

# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.

# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import os

from xklb import usage
from xklb.utils import arggroups, argparse_utils, consts, file_utils, processes


def parse_args():
    parser = argparse_utils.ArgumentParser(usage=usage.mergerfs_cp)
    parser.add_argument("--simulate", "--dry-run", action="store_true", help="Dry run")
    arggroups.clobber(parser)
    arggroups.debug(parser)

    arggroups.paths_or_stdin(parser, destination=True)
    parser.add_argument("destination", help="Destination directory")
    args = parser.parse_args()
    arggroups.args_post(args, parser)
    return args


def get_mergerfs_mounts():
    mounts = []
    with open("/proc/self/mountinfo", "r") as f:
        for line in f:
            values = line.split()
            mountroot, mountpoint = values[3:5]
            separator = values.index("-", 6)
            fstype = values[separator + 1]
            if fstype == "fuse.mergerfs" and mountroot == "/":
                mounts.append(mountpoint.encode().decode("unicode_escape"))
    return mounts


def get_srcmounts(mergerfs_mount):
    import xattr

    ctrlfile = os.path.join(mergerfs_mount, ".mergerfs")
    return xattr.getxattr(ctrlfile, "user.mergerfs.srcmounts").decode().split(":")


def get_destination_mount(destination):
    mergerfs_mounts = get_mergerfs_mounts()

    lcp = ""
    for mergerfs_mount in mergerfs_mounts:
        common_path = os.path.commonpath([mergerfs_mount, destination])
        if len(common_path) > len(lcp):
            lcp = common_path
    return lcp


def mcp_file(args, source, destination):
    found_file = False
    for srcmount in args.srcmounts:
        relative_to_mount = os.path.relpath(source, args.mergerfs_mount)
        original_path = os.path.join(srcmount, relative_to_mount)
        if os.path.exists(original_path):
            found_file = True
            destination = os.path.join(srcmount, os.path.relpath(destination, args.mergerfs_mount))

            src_dest = [original_path, destination]
            if args.simulate:
                print(*args.cp_args, *src_dest)
            else:
                os.makedirs(os.path.dirname(destination), exist_ok=True)
                processes.cmd(*args.cp_args, *src_dest, strict=False, quiet=False, error_verbosity=2)

    if not found_file:
        print(f"Could not find srcmount of {source}")


def cp_args(args):
    cmd_args = ["cp"]
    if args.replace is None:
        cmd_args.append("--interactive")
    elif args.replace is False:
        cmd_args.append("--no-clobber")

    if not consts.PYTEST_RUNNING:
        cmd_args.append("--reflink=always")
    return cmd_args


def mergerfs_cp():
    args = parse_args()

    args.cp_args = cp_args(args)
    args.destination = os.path.realpath(args.destination)
    args.mergerfs_mount = get_destination_mount(args.destination)
    if args.mergerfs_mount == "":
        processes.exit_error("Could not detect any mergerfs mounts")
    args.srcmounts = get_srcmounts(args.mergerfs_mount)

    sources = (os.path.realpath(s) + ("/" if s.endswith("/") else "") for s in args.paths)  # preserve trailing slash
    for source in sources:
        if os.path.isdir(source):
            for p in file_utils.rglob(source, args.ext or None)[0]:
                cp_dest = args.destination
                if not source.endswith(os.sep):  # use BSD behavior
                    cp_dest = os.path.join(cp_dest, os.path.basename(source))
                cp_dest = os.path.join(cp_dest, os.path.dirname(os.path.relpath(p, source)), os.path.basename(p))

                mcp_file(args, p, cp_dest)
        else:
            mcp_file(args, source, args.destination)


if __name__ == "__main__":
    mergerfs_cp()



from xklb import usage
from xklb.utils import arggroups, argparse_utils, file_utils, nums, web
from xklb.utils.consts import DEFAULT_FILE_ROWS_READ_LIMIT
from xklb.utils.log_utils import log
from xklb.utils.printing import print_df, print_series


def parse_args():
    parser = argparse_utils.ArgumentParser(description="Perform EDA on one or more files", usage=usage.eda)

    parser.add_argument("--groupby", "--group-by", "-g", action="store_true")
    arggroups.table_like(parser)

    parser.add_argument("--sort", "-u", default="random()")
    parser.add_argument("--repl", "-r", action="store_true")
    arggroups.debug(parser)

    parser.add_argument(
        "paths",
        metavar="path",
        nargs="+",
        action=argparse_utils.ArgparseArgsOrStdin,
        help="path to one or more files",
    )
    args = parser.parse_args()
    arggroups.args_post(args, parser)

    arggroups.table_like_post(args)

    return args


def df_column_values(df, column_name) -> dict:
    total = len(df)

    null = df[column_name].isnull().sum()
    zero = (df[column_name] == 0).sum()
    empty = (df[column_name] == "").sum()
    values = total - empty - zero - null

    return {
        "values_count": values,
        "null_count": null,
        "zero_count": zero,
        "empty_string_count": empty,
        "column": column_name,
        "null": f"{null} ({nums.percent(null, total):.1f}%)",
        "zero": f"{zero} ({nums.percent(zero, total):.1f}%)",
        "empty_string": f"{empty} ({nums.percent(empty, total):.1f}%)",
        "values": f"{values} ({nums.percent(values, total):.1f}%)",
    }


def print_info(args, df):
    import pandas as pd

    if df.shape == (0, 0):
        print(f"Table [{df.name}] empty")
        return

    if args.end_row is None:
        partial_dataset_msg = ""
    elif args.end_row == DEFAULT_FILE_ROWS_READ_LIMIT:
        partial_dataset_msg = f"(limited by default --end-row {args.end_row})"
    else:
        partial_dataset_msg = f"(limited by --end-row {args.end_row})"
    if args.end_row is not None and args.end_row not in df.shape:
        partial_dataset_msg = ""
    print("### Shape")
    print()
    print(df.shape, partial_dataset_msg)
    print()

    print("### Sample of rows")
    if len(df) > 6:
        print_df(pd.concat([df.head(3), df.tail(3)]))
    else:
        print_df(df.head(6))

    print("### Summary statistics")
    print_df(df.describe())

    converted = df.convert_dtypes()
    same_dtypes = []
    diff_dtypes = []
    for col in df.columns:
        if df.dtypes[col] == converted.dtypes[col]:
            same_dtypes.append((col, df.dtypes[col]))
        else:
            diff_dtypes.append((col, df.dtypes[col], converted.dtypes[col]))
    if len(same_dtypes) > 0:
        print("### Pandas columns with 'original' dtypes")
        same_dtypes = pd.DataFrame(same_dtypes, columns=["column", "dtype"])
        print_df(same_dtypes.set_index("column"))
    if len(diff_dtypes) > 0:
        print("### Pandas columns with 'converted' dtypes")
        diff_dtypes = pd.DataFrame(diff_dtypes, columns=["column", "original_dtype", "converted_dtype"])
        print_df(diff_dtypes.set_index("column"))

    if len(df) > 15:
        numeric_columns = df.select_dtypes("number").columns.to_list()
        if numeric_columns:
            print("### Numerical columns")
            print()
            print("#### Bins")
            print()
            for col in numeric_columns:
                try:
                    bins = pd.cut(df[col], bins=6)
                    print_df(bins.value_counts().sort_index())
                except TypeError:  # putmask: first argument must be an array
                    log.warning("Could not calculate bins for col %s", col)

        categorical_columns = [s for s in df.columns.to_list() if s not in numeric_columns]
        if categorical_columns:
            high_cardinality_cols = set()
            low_cardinality_cols = set()

            print("### Categorical columns")
            print()
            for col in categorical_columns:
                vc = df[col].value_counts()
                vc = vc[vc > (len(df) * 0.005)]
                if len(vc) > 0:
                    low_cardinality_cols.add(col)
                    print(f"#### common values of {col} column")
                    vc = pd.DataFrame({"Count": vc, "Percentage": (vc / len(df)) * 100}).sort_values(
                        by="Count", ascending=False
                    )
                    print_df(vc.head(30))

                    if args.groupby:
                        groups = df.groupby(col).size()
                        groups = groups[groups >= 15]
                        if len(groups) > 0:
                            print(f"#### group by {col}")
                            print_df(df[df[col].isin(groups.index)].groupby(col).describe())

                unique_count = df[col].nunique()
                if unique_count >= (len(df) * 0.2):
                    high_cardinality_cols.add(col)

            med_cardinality_cols = low_cardinality_cols.intersection(high_cardinality_cols)
            low_cardinality_cols = low_cardinality_cols - med_cardinality_cols
            high_cardinality_cols = high_cardinality_cols - med_cardinality_cols

            if high_cardinality_cols:
                print("#### High cardinality (many unique values)")
                print_series(high_cardinality_cols)
            if med_cardinality_cols:
                print("#### Medium cardinality (many unique but also many similar values)")
                print_series(med_cardinality_cols)
            if low_cardinality_cols:
                print("#### Low cardinality (many similar values)")
                print_series(low_cardinality_cols)

    print("### Missing values")
    print()
    nan_col_sums = df.isna().sum()
    print(
        f"{nan_col_sums.sum():,} nulls/NaNs",
        f"({(nan_col_sums.sum() / (df.shape[0] * df.shape[1])):.1%} dataset values missing)",
    )
    print()

    if nan_col_sums.sum():
        no_nas = df.columns[df.notnull().all()]
        if len(no_nas) > 0:
            print(f"#### {len(no_nas)} columns with no missing values")
            print_series(no_nas)

        all_nas = df.columns[df.isnull().all()]
        if len(all_nas) > 0:
            print(f"#### {len(all_nas)} columns with all missing values")
            print_series(all_nas)

        print("#### Value stats")
        column_report = pd.DataFrame(df_column_values(df, col) for col in df.columns).set_index("column")
        column_report = column_report.sort_values(["empty_string_count", "zero_count", "null_count"])
        print_df(column_report[["values", "null", "zero", "empty_string"]])


def file_eda(args, path):
    dfs = file_utils.read_file_to_dataframes(
        path,
        table_name=args.table_name,
        table_index=args.table_index,
        start_row=args.start_row,
        end_row=args.end_row,
        order_by=args.sort,
        encoding=args.encoding,
        mimetype=args.mimetype,
        join_tables=args.join_tables,
    )
    if getattr(args, "repl", False):
        breakpoint()

    for df in dfs:
        print(f"## {path}:{df.name}")
        print_info(args, df)


def eda():
    args = parse_args()
    web.requests_session(args)  # configure session
    for path in args.paths:
        file_eda(args, path)


if __name__ == "__main__":
    eda()



import json, sys

from xklb import usage
from xklb.utils import arggroups, argparse_utils, nums, printing, processes
from xklb.utils.log_utils import log


def parse_utils():
    parser = argparse_utils.ArgumentParser(usage=usage.json_keys_rename)
    arggroups.debug(parser)

    args, unknown_args = parser.parse_known_args()
    arggroups.args_post(args, parser)
    return args, unknown_args


def parse_unknown_args_to_dict(unknown_args):
    kwargs = {}
    key = None
    values = []

    def get_val():
        if len(values) == 1:
            return nums.safe_int_float_str(values[0])
        else:
            return " ".join(values)

    for arg in unknown_args:
        if arg.startswith("--") or arg.startswith("-"):
            if key is not None:
                kwargs[key] = get_val()  # previous values
                values.clear()
            # Process the new key
            key = arg.strip("-").replace("-", "_")
        else:
            values.append(arg)

    if len(values) > 0:
        kwargs[key] = get_val()

    return kwargs


def rename_keys(json_data, key_mapping):
    key_mapping = {old_key: new_key for new_key, old_key in key_mapping.items()}  # swap keys
    keys_to_rename = list(key_mapping.keys())

    new_data = {}
    for key_to_rename in keys_to_rename:
        for old_key in list(json_data.keys()):
            if key_to_rename in old_key.lower():
                new_key = key_mapping[key_to_rename]
                new_data[new_key] = json_data.pop(old_key)
                break

    return new_data


def gen_d(line):
    json_data = json.loads(line)
    if isinstance(json_data, list):
        yield from json_data
    elif isinstance(json_data, dict):
        yield json_data
    else:
        raise TypeError


def json_keys_rename():
    args, unknown_args = parse_utils()

    key_mapping = parse_unknown_args_to_dict(unknown_args)
    if not key_mapping:
        log.error("No data given via arguments")
        raise SystemExit(2)

    print(f"json-keys-rename: Reading from stdin...", file=sys.stderr)
    lines = sys.stdin.readlines()
    if not lines or (len(lines) == 1 and lines[0].strip() == ""):
        processes.exit_error("No data passed in")
    else:
        lines = [s.strip() for s in lines]

    for l in lines:
        for d in gen_d(l):
            renamed_data = rename_keys(d, key_mapping)
            printing.pipe_lines(json.dumps(renamed_data) + "\n")



import argparse, textwrap
from copy import deepcopy
from itertools import groupby

from xklb import usage
from xklb.playback import media_player, media_printer
from xklb.utils import arggroups, argparse_utils, consts, iterables, printing, processes
from xklb.utils.log_utils import log
from xklb.utils.sqlgroups import construct_captions_search_query


def parse_args() -> argparse.Namespace:
    parser = argparse_utils.ArgumentParser(prog="library search", usage=usage.search)

    arggroups.sql_fs(parser)

    arggroups.playback(parser)
    arggroups.post_actions(parser)

    parser.set_defaults(sort=["path", "time"])

    parser.add_argument("--open", "--play", action="store_true")
    parser.add_argument("--overlap", type=int, default=8)
    parser.add_argument("--table", action="store_true")

    parser.set_defaults(print="p")

    arggroups.debug(parser)
    arggroups.database(parser)
    parser.add_argument("search", nargs="*")
    args = parser.parse_intermixed_args()
    arggroups.args_post(args, parser)

    arggroups.sql_fs_post(args)
    arggroups.playback_post(args)
    arggroups.post_actions_post(args)

    return args


def printer(args, captions) -> None:
    captions = iterables.list_dict_filter_bool(captions)
    if not captions:
        processes.no_media_found()

    tbl = deepcopy(captions)
    printing.col_hhmmss(tbl, "time")

    if args.print == "p":
        print(f"{len(captions)} captions")
        for path, path_group in groupby(tbl, key=lambda x: x["path"]):
            path_group = list(path_group)
            title = path_group[0].get("title")
            print(" - ".join(iterables.concat(title, path)))
            for caption in path_group:
                for line in textwrap.wrap(caption["text"], subsequent_indent=" " * 9, initial_indent=f"{caption['time']} ", width=consts.TERMINAL_SIZE.columns - 2):  # type: ignore
                    print(line)
            print()
    else:
        media_printer.media_printer(args, captions, units="captions")


def merge_captions(args, captions):
    def get_end(caption):
        return caption["time"] + (len(caption["text"]) / 4.2 / 220 * 60)

    merged_captions = []
    for path, group in groupby(
        captions,
        key=lambda x: x["path"],
    ):  # group by only does contiguous items with the same key
        group = list(group)
        merged_group = {"path": path, "title": group[0]["title"], "time": group[0]["time"], "end": get_end(group[0]), "text": group[0]["text"]}  # type: ignore
        for i in range(1, len(group)):
            end = get_end(group[i])

            if (
                abs(group[i]["time"] - merged_group["end"]) <= args.overlap  # type: ignore
                or abs(group[i]["time"] - merged_group["time"]) <= args.overlap  # type: ignore
            ):
                merged_group["end"] = end
                if group[i]["text"] not in merged_group["text"]:  # type: ignore
                    merged_group["text"] += ". " + group[i]["text"]  # type: ignore
            else:
                merged_captions.append(merged_group)
                merged_group = {
                    "path": path,
                    "time": group[i]["time"],  # type: ignore
                    "end": end,
                    "text": group[i]["text"],  # type: ignore
                }
        merged_captions.append(merged_group)

    return merged_captions


def search() -> None:
    args = parse_args()
    query, bindings = construct_captions_search_query(args)
    captions = list(args.db.query(query, bindings))
    merged_captions = merge_captions(args, captions)

    if args.open:
        pl = media_player.MediaPrefetcher(args, merged_captions)
        pl.fetch()
        while pl.remaining:
            d = pl.get_m()
            if d:
                print(d["text"])
                m = args.db.pop_dict("select * from media where path = ?", [d["path"]])
                m["player"].extend([f'--start={d["time"] - 2}', f'--end={int(d["end"] + 1.5)}'])
                r = media_player.single_player(args, m)
                if r.returncode != 0:
                    log.warning("Player exited with code %s", r.returncode)
                    if args.ignore_errors:
                        return
                    else:
                        raise SystemExit(r.returncode)
    else:
        printer(args, merged_captions)


if __name__ == "__main__":
    search()



import argparse, json, math, os, re, sys
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from functools import partial
from multiprocessing import TimeoutError as mp_TimeoutError
from pathlib import Path
from shutil import which
from timeit import default_timer as timer

from xklb import usage
from xklb.createdb import av
from xklb.files import sample_hash
from xklb.folders import rel_mv
from xklb.mediadb import db_media, db_playlists, playlists
from xklb.mediafiles import process_ffmpeg, process_image
from xklb.utils import (
    arg_utils,
    arggroups,
    argparse_utils,
    consts,
    db_utils,
    file_utils,
    iterables,
    nums,
    objects,
    path_utils,
    printing,
    processes,
    strings,
)
from xklb.utils.consts import SC, DBType
from xklb.utils.log_utils import log

REGEX_SENTENCE_ENDS = re.compile(r";|,|\.|\*|\n|\t")


def parse_args(action, usage):
    parser = argparse_utils.ArgumentParser(prog="library " + action, usage=usage)
    arggroups.db_profiles(parser)
    arggroups.simulate(parser)

    parser.add_argument("--exclude", "-E", nargs="+", action="extend", default=[])

    parser.add_argument("--ocr", "--OCR", action="store_true")
    parser.add_argument("--speech-recognition", "--speech", action="store_true")
    parser.add_argument("--scan-subtitles", "--scan-subtitle", action="store_true")

    parser.add_argument("--hash", action="store_true")

    parser.add_argument("--process", action="store_true")
    arggroups.process_ffmpeg(parser)

    parser.add_argument("--check-corrupt", "--check-corruption", action="store_true")
    arggroups.media_check(parser)
    parser.set_defaults(gap="0.10")

    parser.add_argument(
        "--force", "-f", action="store_true", help="Mark all subpath files as deleted if no files found"
    )
    parser.add_argument("--move")

    arggroups.debug(parser)

    arggroups.database(parser)
    if action == SC.fs_add:
        arggroups.paths_or_stdin(parser)
    args = parser.parse_intermixed_args()
    arggroups.args_post(args, parser, create_db=action == SC.fs_add)

    if not args.profiles:
        args.profiles = [DBType.video]

    if args.move:
        args.move = Path(args.move).expanduser().resolve()

    if hasattr(args, "paths"):
        args.paths = iterables.conform(args.paths)

    if not which("ffprobe") and (DBType.audio in args.profiles or DBType.video in args.profiles):
        log.error("ffmpeg is not installed. Install it with your package manager.")
        raise SystemExit(3)

    arggroups.process_ffmpeg_post(args)
    arggroups.media_check_post(args)

    return args


def munge_book_tags(path) -> dict:
    try:
        import textract  # type: ignore
    except ModuleNotFoundError:
        print(
            "textract is required for text database creation: pip install textract; sudo dnf install libxml2-devel libxslt-devel antiword unrtf poppler-utils tesseract sox-plugins-nonfree sox libjpeg-devel swig",
        )
        raise
    try:
        tags = textract.process(path, language=os.getenv("TESSERACT_LANGUAGE"))
        tags = REGEX_SENTENCE_ENDS.split(tags.decode())
    except Exception as e:
        log.warning(e)
        log.error(f"Failed reading text. {path}")
        tags = []
    return {"tags": strings.combine(tags)}


munge_book_tags_fast = processes.with_timeout(70)(munge_book_tags)
munge_book_tags_slow = processes.with_timeout(350)(munge_book_tags)


def extract_metadata(mp_args, path) -> dict[str, str | int | None] | None:
    try:
        path.encode()
    except UnicodeEncodeError:
        log.error("Could not encode file path as UTF-8. Skipping %s", path)
        return None

    try:
        stat = os.stat(path, follow_symlinks=False)
    except FileNotFoundError:
        return None
    except OSError:
        log.exception("OSError: possible filesystem corruption; check dmesg. %s", path)
        return None
    except Exception as e:
        log.error(f"%s {path}", e)
        return None

    media = {
        "path": path,
        "size": stat.st_size,
        "type": file_utils.mimetype(path),
        "time_created": int(stat.st_ctime),
        "time_modified": int(stat.st_mtime) or consts.now(),
        "time_downloaded": consts.APPLICATION_START,
        "time_deleted": 0,
    }

    ext = path.rsplit(".", 1)[-1].lower()
    is_scan_all_files = getattr(mp_args, "scan_all_files", False)

    if media["type"] == "directory":
        return None

    if media["size"] == 0 or not Path(path).exists():
        return media

    if objects.is_profile(mp_args, DBType.audio) and (ext in consts.AUDIO_ONLY_EXTENSIONS or is_scan_all_files):
        media |= av.munge_av_tags(mp_args, path)
    elif objects.is_profile(mp_args, DBType.video) and (ext in consts.VIDEO_EXTENSIONS or is_scan_all_files):
        media |= av.munge_av_tags(mp_args, path)

    if not Path(path).exists():  # av.munge_av_tags might delete if unplayable or corruption exceeds threshold
        return media

    text_exts = consts.TEXTRACT_EXTENSIONS
    if mp_args.ocr:
        text_exts |= consts.OCR_EXTENSIONS
    if mp_args.speech_recognition:
        text_exts |= consts.SPEECH_RECOGNITION_EXTENSIONS
    if objects.is_profile(mp_args, DBType.text) and (ext in text_exts or is_scan_all_files):
        try:
            start = timer()
            if any([mp_args.ocr, mp_args.speech_recognition]):
                media |= munge_book_tags_slow(path)
            else:
                media |= munge_book_tags_fast(path)
        except mp_TimeoutError:
            log.warning(f"Timed out trying to read file. {path}")
        else:
            log.debug(f"{timer()-start} {path}")

    if getattr(mp_args, "hash", False) and media["type"] != "directory" and media["size"] > 0:
        media["hash"] = sample_hash.sample_hash_file(path)

    if getattr(mp_args, "move", False) and not file_utils.is_file_open(path):
        dest_path = rel_mv.gen_rel_path(path, mp_args.move)
        dest_path = path_utils.clean_path(bytes(dest_path))
        file_utils.rename_move_file(path, dest_path, simulate=mp_args.simulate)
        path = media["path"] = dest_path

    if getattr(mp_args, "process", False):
        if objects.is_profile(mp_args, DBType.audio) and Path(path).suffix not in [".opus", ".mka"]:
            result = process_ffmpeg.process_path(
                mp_args,
                path,
                split_longer_than=2160 if mp_args.split_longer_than is None and "audiobook" in path.lower() else None,
            )
            if result is None:
                return None
            path = media["path"] = str(result)
        elif objects.is_profile(mp_args, DBType.video) and Path(path).suffix not in [".av1.mkv"]:
            result = process_ffmpeg.process_path(mp_args, path)
            if result is None:
                return None
            path = media["path"] = str(result)
        elif objects.is_profile(mp_args, DBType.image) and Path(path).suffix not in [".avif", ".avifs"]:
            result = process_image.process_path(mp_args, path)
            if result is None:
                return None
            path = media["path"] = str(result)

    return media


def clean_up_temp_dirs():
    temp_subs_path = Path(consts.SUB_TEMP_DIR)
    if temp_subs_path.exists():
        for p in temp_subs_path.glob("*.srt"):
            p.unlink()


def pop_substring_keys(e, key_substring):
    values = []
    for k in list(e.keys()):
        if key_substring in k:
            values.append(e.pop(k))
    return values


def munge_image_tags(m: dict, e: dict) -> dict:
    chroma_subsample = nums.safe_int(
        iterables.safe_unpack(
            e.pop("File:YCbCrSubSampling", None),
            *pop_substring_keys(e, "YCbCrSubSampling"),
        )
    )
    if chroma_subsample == 0:
        chroma_subsample = None

    unit_x = iterables.safe_unpack(
        *pop_substring_keys(e, "XResolution"),
    )
    unit_y = iterables.safe_unpack(
        *pop_substring_keys(e, "YResolution"),
    )
    unit = iterables.safe_unpack(
        *pop_substring_keys(e, "ResolutionUnit"),
    )
    if unit == 0:
        unit = None
        unit_x = None if unit_x == 1 else unit_x
        unit_y = None if unit_y == 1 else unit_y

    m = {
        **m,
        "orientation": iterables.safe_unpack(
            *pop_substring_keys(e, "Orientation"),
        ),
        "width": iterables.safe_unpack(
            e.pop("File:ImageWidth", None),
            e.pop("Composite:ImageWidth", None),
            e.pop("EXIF:ImageWidth", None),
            e.pop("EXIF:ExifImageWidth", None),
            e.pop("PNG:ImageWidth", None),
            *pop_substring_keys(e, "ImageWidth"),
        ),
        "height": iterables.safe_unpack(
            e.pop("File:ImageHeight", None),
            e.pop("Composite:ImageHeight", None),
            e.pop("EXIF:ImageHeight", None),
            e.pop("EXIF:ExifImageHeight", None),
            e.pop("PNG:ImageHeight", None),
            *pop_substring_keys(e, "ImageHeight"),
        ),
        "chroma_subsample": chroma_subsample,
        "color_depth": iterables.safe_unpack(
            *pop_substring_keys(e, "ColorResolutionDepth"),
        ),
        "color_background": iterables.safe_unpack(
            *pop_substring_keys(e, "BackgroundColor"),
        ),
        "color_transparent": iterables.safe_unpack(
            *pop_substring_keys(e, "TransparentColor"),
        ),
        "longitude": iterables.safe_unpack(
            e.pop("Composite:GPSLongitude", None),
            *pop_substring_keys(e, "GPSLongitude"),
        ),
        "latitude": iterables.safe_unpack(
            e.pop("Composite:GPSLatitude", None),
            *pop_substring_keys(e, "GPSLatitude"),
        ),
        "focal_length": iterables.safe_unpack(
            e.pop("Composite:Lens", None),
            *pop_substring_keys(e, "Lens35efl"),
        ),
        "unit": unit,
        "unit_x": unit_x,
        "unit_y": unit_y,
        "exiftool_warning": strings.combine(*pop_substring_keys(e, "ExifTool:Warning")),
        "tags": strings.combine(
            *pop_substring_keys(e, "Headline"),
            *pop_substring_keys(e, "Title"),
            *pop_substring_keys(e, "ImageDescription"),
            *pop_substring_keys(e, "Caption"),
            *pop_substring_keys(e, "Artist"),
            *pop_substring_keys(e, "By-line"),
            *pop_substring_keys(e, "Credit"),
            *pop_substring_keys(e, "DocumentNotes"),
            *pop_substring_keys(e, "URL_List"),
            *pop_substring_keys(e, "Keywords"),
            *pop_substring_keys(e, "Make"),
            *pop_substring_keys(e, "Model"),
            *pop_substring_keys(e, "LensID"),
            *pop_substring_keys(e, "Creator"),
            *pop_substring_keys(e, "Software"),
        ),
    }

    return m


def extract_image_metadata_chunk(metadata: list[dict]) -> list[dict]:
    try:
        import exiftool
    except ModuleNotFoundError:
        print(
            "exiftool and PyExifTool are required for image database creation: sudo dnf install perl-Image-ExifTool && pip install PyExifTool",
        )
        raise

    chunk_paths = [d["path"] for d in metadata]
    try:
        with exiftool.ExifToolHelper() as et:
            exif = et.get_metadata(chunk_paths)
    except exiftool.exceptions.ExifToolExecuteError:
        log.exception("exifTool failed executing get_metadata %s", metadata)
        return metadata

    exif_enriched = []
    for m, e in zip(metadata, exif):
        assert m["path"] == e.pop("SourceFile")

        try:
            m = munge_image_tags(m, e)
        except Exception as e:
            log.error("[%s]: %s", m["path"], e)
            # continue ?
        exif_enriched.append(m)

    return exif_enriched


def extract_chunk(args, media) -> None:
    if objects.is_profile(args, DBType.image):
        media = extract_image_metadata_chunk(media)

    if args.scan_subtitles:
        clean_up_temp_dirs()

    captions = []
    for d in media:
        caption = {}
        caption["path"] = d["path"]

        caption["chapters"] = d.pop("chapters", None) or []
        caption["subtitles"] = d.pop("subtitles", None) or []

        tags = d.pop("tags", None) or ""
        description = d.pop("description", None) or ""
        if description:
            tags += "\n" + description
        if tags:
            caption["captions_t0"] = {"time": 0, "text": tags}

        captions.append(caption)

    media = iterables.list_dict_filter_bool(media)
    media = [{"playlists_id": args.playlists_id, **d} for d in media]
    args.db["media"].insert_all(media, pk="id", alter=True, replace=True)

    for d in captions:
        media_id = args.db.pop("select id from media where path = ?", [d["path"]])
        if len(d["chapters"]) > 0:
            args.db["captions"].insert_all([{**d, "media_id": media_id} for d in d["chapters"]], alter=True)
        if len(d["subtitles"]) > 0:
            args.db["captions"].insert_all([{**d, "media_id": media_id} for d in d["subtitles"]], alter=True)
        if d.get("caption_t0"):
            args.db["captions"].insert({**d["caption_t0"], "media_id": media_id}, alter=True)


def mark_media_undeleted(args, paths) -> int:
    paths = iterables.conform(paths)

    modified_row_count = 0
    if paths:
        df_chunked = iterables.chunks(paths, consts.SQLITE_PARAM_LIMIT)
        for chunk_paths in df_chunked:
            with args.db.conn:
                cursor = args.db.conn.execute(
                    """update media
                    set time_deleted=0
                    where path in ("""
                    + ",".join(["?"] * len(chunk_paths))
                    + ")",
                    (*chunk_paths,),
                )
                modified_row_count += cursor.rowcount

    return modified_row_count


def find_new_files(args, path) -> list[str]:
    if path.is_file():
        scanned_set = {str(path)}
    else:
        for s in args.profiles:
            if getattr(DBType, s, None) is None:
                msg = f"fs_extract for profile {s}"
                raise NotImplementedError(msg)

        exts = args.ext
        if not exts:
            exts = set()
            if args.scan_all_files or DBType.filesystem in args.profiles:
                exts = None
            else:
                if DBType.audio in args.profiles:
                    exts |= consts.AUDIO_ONLY_EXTENSIONS
                if DBType.video in args.profiles:
                    exts |= consts.VIDEO_EXTENSIONS

                if DBType.image in args.profiles:
                    exts |= consts.IMAGE_EXTENSIONS

                if DBType.text in args.profiles:
                    exts |= consts.TEXTRACT_EXTENSIONS
                if args.ocr:
                    exts |= consts.OCR_EXTENSIONS
                if args.speech_recognition:
                    exts |= consts.SPEECH_RECOGNITION_EXTENSIONS

        scanned_set = file_utils.rglob(path, exts or None, args.exclude)[0]

    m_columns = db_utils.columns(args, "media")

    try:
        deleted_set = {
            d["path"]
            for d in args.db.query(
                f"""select path from media
                where 1=1
                    and time_deleted > 0
                    and path like '{path}%'
                    {'AND time_downloaded > 0' if 'time_downloaded' in m_columns else ''}
                """,
            )
        }
    except Exception as e:
        log.debug(e)
    else:
        undeleted_files = list(deleted_set.intersection(scanned_set))
        undeleted_count = mark_media_undeleted(args, undeleted_files)
        if undeleted_count > 0:
            print(f"[{path}] Marking", undeleted_count, "metadata records as undeleted")

    try:
        existing_set = {
            d["path"]
            for d in args.db.query(
                f"""select path from media
                where 1=1
                    and path like ?
                    and coalesce(time_deleted, 0) = 0
                    {'AND time_downloaded > 0' if 'time_downloaded' in m_columns else ''}
                """,
                [str(path) + "%"],
            )
        }
    except Exception as e:
        log.debug(e)
        new_files = list(scanned_set)
    else:
        new_files = list(scanned_set - existing_set)

        deleted_files = list(existing_set - scanned_set)
        if not scanned_set and len(deleted_files) >= len(existing_set) and not args.force:
            print(f"[{path}] Path empty or device not mounted. Rerun with -f to mark all subpaths as deleted.")
            return []  # if path not mounted or all files deleted
        deleted_count = db_media.mark_media_deleted(args, deleted_files)
        if deleted_count > 0:
            print(f"[{path}] Marking", deleted_count, "orphaned metadata records as deleted")

    new_files.sort(key=len, reverse=True)
    return new_files


def scan_path(args, path_str: str) -> int:
    path = Path(path_str).expanduser().resolve()
    if not path.exists():
        print(f"[{path}] Path does not exist")
        if args.force:
            playlists.delete_playlists(args, [str(path)])
        return 0

    n_jobs = None
    if args.verbose >= consts.LOG_DEBUG:
        n_jobs = 1
    elif args.threads != -1:
        n_jobs = args.threads

    threadsafe = [DBType.audio, DBType.video, DBType.filesystem]

    info = {
        "extractor_key": "Local",
        "extractor_config": args.extractor_config,
        "time_deleted": 0,
    }
    args.playlists_id = db_playlists.add(args, str(path), info, check_subpath=True)

    print(f"[{path}] Building file list...")
    new_files = find_new_files(args, path)
    if new_files:
        print(f"[{path}] Adding {len(new_files)} new media")
        # log.debug(new_files)

        if DBType.text in args.profiles:
            batch_count = int(os.cpu_count() or 4)
        elif DBType.image in args.profiles:
            batch_count = consts.SQLITE_PARAM_LIMIT // 20
        else:
            batch_count = consts.SQLITE_PARAM_LIMIT // 100
        chunks_count = math.ceil(len(new_files) / batch_count)
        files_chunked = iterables.chunks(new_files, batch_count)

        if all(s in threadsafe for s in args.profiles):
            pool_fn = ThreadPoolExecutor
        else:
            pool_fn = ProcessPoolExecutor

        with pool_fn(n_jobs) as parallel:
            for idx, chunk_paths in enumerate(files_chunked):
                percent = ((batch_count * idx) + len(chunk_paths)) / len(new_files) * 100
                printing.print_overwrite(
                    f"[{path}] Extracting metadata {percent:3.1f}% (chunk {idx + 1} of {chunks_count})"
                )

                mp_args = argparse.Namespace(
                    playlist_path=path, **{k: v for k, v in args.__dict__.items() if k not in {"db"}}
                )
                metadata = parallel.map(partial(extract_metadata, mp_args), chunk_paths)
                metadata = list(filter(None, metadata))
                extract_chunk(args, metadata)
            print()

    return len(new_files)


def extractor(args, paths) -> None:
    new_files = 0
    for path in paths:
        new_files += scan_path(args, path)

    log.info("Imported %s paths", new_files)

    if not args.db["media"].detect_fts() or new_files > 100000:
        db_utils.optimize(args)


def fs_add(args=None) -> None:
    if args:
        sys.argv = ["lb", *args]

    args = parse_args(SC.fs_add, usage.fs_add)
    extractor(args, args.paths)


def fs_update(args=None) -> None:
    if args:
        sys.argv = ["lb", *args]

    args = parse_args(SC.fs_update, usage.fs_update)

    fs_playlists = list(
        args.db.query(
            f"""
            SELECT *
            FROM playlists
            WHERE extractor_key = 'Local'
            ORDER BY
                length(path)-length(REPLACE(path, '{os.sep}', '')) desc
                , path
            """,
        ),
    )

    for playlist in fs_playlists:
        extractor_config = json.loads(playlist.get("extractor_config") or "{}")
        args_env = arg_utils.override_config(args, extractor_config)

        extractor(args_env, [playlist["path"]])



import argparse, json, operator, random
from pathlib import Path

from xklb.utils import consts, file_utils, iterables
from xklb.utils.consts import SC


def gen_paths(args):
    if args.from_file:
        for path in args.paths:
            with open(path, "r") as f:
                for line in f:
                    line = line.rstrip("\n")
                    if line.strip():
                        if args.from_json:
                            json_data = json.loads(line)
                            if isinstance(json_data, list):
                                yield from (d["path"] for d in json_data)
                            elif isinstance(json_data, dict):
                                yield json_data["path"]
                            else:
                                raise TypeError
                        else:
                            yield line
    elif args.from_json:
        for path in args.paths:
            json_data = json.loads(path)
            if isinstance(json_data, list):
                yield from (d["path"] for d in json_data)
            elif isinstance(json_data, dict):
                yield json_data["path"]
            else:
                raise TypeError
    else:
        is_large = len(args.paths) > 1000
        for path in args.paths:
            if path.strip():
                if is_large:
                    yield path
                else:
                    p = Path(path)
                    if p.is_dir():
                        yield from file_utils.rglob(str(p), args.ext or None, getattr(args, "exclude", None))[0]
                    else:
                        yield path


def gen_d(args):
    if args.from_file:
        for path in args.paths:
            with open(path, "r") as f:
                for line in f:
                    line = line.rstrip("\n")
                    if line.strip():
                        if args.from_json:
                            json_data = json.loads(line)
                            if isinstance(json_data, list):
                                yield from json_data
                            elif isinstance(json_data, dict):
                                yield json_data
                            else:
                                raise TypeError
                        else:
                            yield {"path": line}
    elif args.from_json:
        for path in args.paths:
            json_data = json.loads(path)
            if isinstance(json_data, list):
                yield from json_data
            elif isinstance(json_data, dict):
                yield json_data
            else:
                raise TypeError
    else:
        for path in args.paths:
            if path.strip():
                p = Path(path)
                if p.is_dir():
                    for sp in file_utils.rglob(str(p), args.ext or None, getattr(args, "exclude", None))[0]:
                        yield {"path": sp}
                else:
                    yield {"path": path}


def override_sort(sort_expression: str) -> str:
    def year_month_sql(var):
        return f"cast(strftime('%Y%m', datetime({var}, 'unixepoch')) as int)"

    def year_month_day_sql(var):
        return f"cast(strftime('%Y%m%d', datetime({var}, 'unixepoch')) as int)"

    return (
        sort_expression.replace("month_created", year_month_sql("time_created"))
        .replace("month_modified", year_month_sql("time_modified"))
        .replace("date_created", year_month_day_sql("time_created"))
        .replace("date_modified", year_month_day_sql("time_modified"))
        .replace("random()", "random")
        .replace("random", "random()")
        .replace("priorityfast", "ntile(1000) over (order by size) desc, duration")
        .replace("priority", "ntile(1000) over (order by size/duration) desc")
        .replace("bitrate", "size/duration desc")
    )


def parse_ambiguous_sort(sort):
    combined_sort = []
    for s in iterables.flatten([s.split(",") for s in sort]):
        if s.strip() in ["asc", "desc"] and combined_sort:
            combined_sort[-1] += " " + s.strip()
        else:
            combined_sort.append(s.strip())
    return combined_sort


def parse_args_sort(args, columns, table_prefix="m.") -> tuple[str, list[str]]:
    sort_list = []
    select_list = []
    if args.sort:
        combined_sort = parse_ambiguous_sort(args.sort)

        for s in combined_sort:
            if s.startswith("same-"):
                var = s[len("same-") :]
                direction = "DESC"
                if var.lower().endswith((" asc", " desc")):
                    var, direction = var.split(" ")

                select_list.append(
                    f"CASE WHEN {var} IS NULL THEN NULL ELSE COUNT(*) OVER (PARTITION BY {var}) END AS same_{var}_count",
                )
                sort_list.append(f"same_{var}_count {direction}")
            else:
                sort_list.append(s)

    # switching between videos with and without subs is annoying
    subtitle_count = "=0"
    if random.random() < getattr(args, "subtitle_mix", consts.DEFAULT_SUBTITLE_MIX):
        # bias slightly toward videos without subtitles
        subtitle_count = ">0"

    sorts = [
        "random" if getattr(args, "random", False) else None,
        "rank" if sort_list and "rank" in sort_list else None,
        "video_count > 0 desc" if "video_count" in columns and args.action == SC.watch else None,
        "audio_count > 0 desc" if "audio_count" in columns else None,
        table_prefix + 'path like "http%"',
        "width < height desc" if "width" in columns and getattr(args, "portrait", False) else None,
        (
            f"subtitle_count {subtitle_count} desc"
            if "subtitle_count" in columns
            and args.action == SC.watch
            and not any(
                [
                    args.print,
                    consts.PYTEST_RUNNING,
                    "subtitle_count" in " ".join(args.where),
                    args.limit != consts.DEFAULT_PLAY_QUEUE,
                ],
            )
            else None
        ),
        *(sort_list or []),
        "play_count" if args.action in (SC.media, SC.listen, SC.watch) else None,
        "time_last_played" if args.action in (SC.media, SC.listen, SC.watch) else None,
        "duration desc" if args.action in (SC.media, SC.listen, SC.watch) and args.include else None,
        "size desc" if args.action in (SC.media, SC.listen, SC.watch) and args.include else None,
        table_prefix + "title IS NOT NULL desc" if "title" in columns else None,
        table_prefix + "path",
    ]

    sort = list(filter(bool, sorts))
    sort = [override_sort(s) for s in sort]
    sort = ",".join(sort)
    return sort.replace(",,", ","), select_list


def parse_args_limit(args):
    if not args.limit:
        if not any(
            [
                args.print and len(args.print.replace("p", "")) > 0,
                getattr(args, "partial", False),
                getattr(args, "lower", False),
                getattr(args, "upper", False),
            ],
        ):
            if args.action in (SC.media, SC.listen, SC.watch, SC.read):
                args.limit = consts.DEFAULT_PLAY_QUEUE
            elif args.action in (SC.view):
                args.limit = consts.DEFAULT_PLAY_QUEUE * 4
            elif args.action in (SC.links_open):
                args.limit = consts.MANY_LINKS - 1
            elif args.action in (SC.download):
                args.limit = consts.DEFAULT_PLAY_QUEUE * 60
    elif args.limit.lower() in ("inf", "all"):
        args.limit = None
    else:
        args.limit = int(args.limit)


def split_folder_glob(s):
    p = Path(s).resolve()

    if "*" not in s and not p.exists():
        p.mkdir(parents=True, exist_ok=True)

    if p.is_dir():
        return p, "*"
    return p.parent, p.name


def override_config(args, extractor_config):
    defaults = getattr(args, "defaults", None) or {}
    overridden_args = {k: v for k, v in args.__dict__.items() if defaults.get(k) != v}
    args_env = argparse.Namespace(
        **{**defaults, **(extractor_config.get("extractor_config") or {}), **extractor_config, **overridden_args}
    )
    return args_env


def kwargs_overwrite(namespace, kwargs):
    namespace_dict = vars(namespace)
    namespace_dict.update(kwargs)
    return argparse.Namespace(**namespace_dict)


ops = {"<": operator.lt, "<=": operator.le, "==": operator.eq, "!=": operator.ne, ">=": operator.ge, ">": operator.gt}


def cmp(arg1, op, arg2):
    operation = ops.get(op)
    return operation(arg1, arg2)  # type: ignore



import argparse

import humanize

from xklb import usage
from xklb.folders.similar_folders import cluster_folders, map_and_name
from xklb.utils import arg_utils, arggroups, argparse_utils, file_utils, nums, printing
from xklb.utils.log_utils import log


def parse_args():
    parser = argparse_utils.ArgumentParser(usage=usage.similar_files)
    arggroups.cluster(parser)

    parser.add_argument("--small", "--reverse", action="store_true")
    parser.add_argument("--only-duplicates", action="store_true")
    parser.add_argument("--only-originals", action="store_true")

    parser.add_argument("--full-path", action="store_true", help="Cluster using full path instead of only file name")
    parser.add_argument("--estimated-duplicates", "--dupes", type=float)

    parser.add_argument("--durations-delta", "--duration-delta", type=float, default=10.0)
    parser.add_argument("--sizes-delta", "--size-delta", type=float, default=10.0)

    parser.add_argument("--filter-names", action=argparse.BooleanOptionalAction, default=True)
    parser.add_argument("--filter-sizes", action=argparse.BooleanOptionalAction, default=True)
    parser.add_argument("--filter-durations", action=argparse.BooleanOptionalAction, default=True)
    arggroups.debug(parser)

    arggroups.paths_or_stdin(parser)
    args = parser.parse_args()
    arggroups.args_post(args, parser)

    if not args.filter_names and not args.filter_sizes and not args.filter_durations:
        print("Nothing to do")
        raise NotImplementedError

    return args


def is_same_size_group(args, m0, m):
    bools = []

    if args.filter_sizes:
        size_diff = nums.percentage_difference(m0["size"], m["size"])
        bools.append(size_diff < args.sizes_delta)

    if args.filter_durations and m.get("duration"):
        duration_diff = nums.percentage_difference(m0["duration"], m["duration"])
        bools.append(duration_diff < args.durations_delta)

    return all(bools)


def cluster_by_size(args, media):
    group_id = 0
    temp_groups = []
    media_groups = []
    for m in media:
        grouped = False
        for i, group in enumerate(temp_groups):
            m0 = group[0]
            if is_same_size_group(args, m0, m):
                group.append(m)
                media_groups.append(i)
                grouped = True
                break

        if not grouped:
            temp_groups.append([m])
            media_groups.append(group_id)
            group_id += 1

    assert len(media_groups) == len(media)
    return media_groups


def filter_group_by_size(args, group):
    media = group["grouped_paths"]

    output = []
    m0 = media[0]
    for m in media[1:]:
        if is_same_size_group(args, m0, m):
            output.append(m)

    if output:
        group["grouped_paths"] = [m0] + output
        return group
    else:
        return None


def filter_groups_by_size(args, groups):
    groups = [filter_group_by_size(args, group) for group in groups]
    groups = [group for group in groups if group]
    return groups


def similar_files():
    args = parse_args()
    media = list(arg_utils.gen_d(args))
    media = [d if "size" in d else file_utils.get_filesize(d) for d in media]

    groups: list[dict] = []
    if args.filter_sizes or args.filter_durations:
        clusters = cluster_by_size(args, media)
        groups = map_and_name(media, clusters)
        log.info("file size/duration clustering sorted %s files into %s groups", len(media), len(groups))
        single_file_groups = [d for d in groups if len(d["grouped_paths"]) == 1]
        groups = [d for d in groups if len(d["grouped_paths"]) > 1]
        log.info("Filtered out %s single-file groups", len(single_file_groups))
        log.debug(single_file_groups)

        if args.filter_names:
            media = [d for group in groups for d in group["grouped_paths"]]

    if args.filter_names:
        groups = cluster_folders(args, media)
        groups = sorted(groups, key=lambda d: (-len(d["grouped_paths"]), -len(d["common_path"])))
        log.info("Name clustering sorted %s files into %s groups", len(media), len(groups))

        if args.filter_sizes or args.filter_durations:
            prev_count = len(groups)
            groups = filter_groups_by_size(args, groups)  # effectively a second pass
            log.info("(2nd pass) group size/duration filtering removed %s groups", prev_count - len(groups))

    if not args.only_originals and not args.only_duplicates:
        print("Duplicate groups:")

    for group in groups:
        media = group["grouped_paths"]
        media = sorted(media, key=lambda d: d["size"], reverse=not args.small)

        if args.only_originals:
            media = media[:1]
        if args.only_duplicates:
            media = media[1:]

        if "f" in args.print:
            for d in media:
                printing.pipe_print(d["path"])
        else:
            printing.table(
                [
                    {
                        f'group {group["common_path"]}': d["path"],
                        "size": humanize.naturalsize(d["size"], binary=True),
                        "duration": printing.human_duration(d.get("duration")),
                    }
                    for d in media
                ]
            )

        if not args.only_originals and not args.only_duplicates:
            print()

    if not args.only_originals and not args.only_duplicates:
        print(len(groups), "groups found")



import argparse, datetime, functools, os, re, tempfile, time, urllib.error, urllib.parse, urllib.request
from email.message import Message
from pathlib import Path
from shutil import which
from urllib.parse import parse_qs, parse_qsl, quote, unquote, urlencode, urljoin, urlparse, urlunparse

import requests
from idna import decode as puny_decode
from idna import encode as puny_encode

from xklb.utils import consts, db_utils, iterables, nums, path_utils, pd_utils, strings, web
from xklb.utils.log_utils import clamp_index, log

headers = {"User-Agent": "Mozilla/5.0 (X11; Linux x86_64; rv:121.0) Gecko/20100101 Firefox/121.0"}
session = None


def _get_retry_adapter(max_retries):
    import requests.adapters

    retry = requests.adapters.Retry(
        total=max_retries,
        connect=max_retries,
        read=max_retries,
        status=max_retries // 2,
        other=1,
        redirect=4,
        raise_on_redirect=False,
        backoff_factor=3,
        backoff_jitter=2,
        backoff_max=22 * 60,
        status_forcelist=[
            104,
            413,
            429,
            500,
            502,
            503,
            504,
            522,
        ],
    )

    return requests.adapters.HTTPAdapter(max_retries=retry)


def parse_cookies_from_browser(input_str):
    from yt_dlp.cookies import SUPPORTED_BROWSERS, SUPPORTED_KEYRINGS

    # lifted from yt_dlp to have a compatible interface
    container = None
    mobj = re.fullmatch(
        r"""(?x)
        (?P<name>[^+:]+)
        (?:\s*\+\s*(?P<keyring>[^:]+))?
        (?:\s*:\s*(?!:)(?P<profile>.+?))?
        (?:\s*::\s*(?P<container>.+))?
    """,
        input_str,
    )
    if mobj is None:
        raise ValueError(f"invalid cookies from browser arguments: {input_str}")
    browser_name, keyring, profile, container = mobj.group("name", "keyring", "profile", "container")
    browser_name = browser_name.lower()
    if browser_name not in SUPPORTED_BROWSERS:
        raise ValueError(
            f'unsupported browser specified for cookies: "{browser_name}". '
            f'Supported browsers are: {", ".join(sorted(SUPPORTED_BROWSERS))}'
        )
    if keyring is not None:
        keyring = keyring.upper()
        if keyring not in SUPPORTED_KEYRINGS:
            raise ValueError(
                f'unsupported keyring specified for cookies: "{keyring}". '
                f'Supported keyrings are: {", ".join(sorted(SUPPORTED_KEYRINGS))}'
            )
    return (browser_name, profile, keyring, container)


def requests_session(args=argparse.Namespace()):
    global session  # TODO: maybe run_once similar to log_utils.log

    if session is None:
        import requests

        http_max_retries = getattr(args, "http_max_retries", None) or 8
        cookie_file = getattr(args, "cookies", None)
        cookies_from_browser = getattr(args, "cookies_from_browser", None)

        session = requests.Session()
        session.mount("http://", _get_retry_adapter(http_max_retries))
        session.mount("https://", _get_retry_adapter(http_max_retries))
        session.request = functools.partial(session.request, headers=headers, timeout=(5, 45))  # type: ignore

        if getattr(args, "allow_insecure", False):
            from urllib3.exceptions import InsecureRequestWarning

            requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)  # type: ignore
            session.verify = False

        if cookie_file or cookies_from_browser:
            from yt_dlp.cookies import load_cookies

            if cookies_from_browser:
                cookies_from_browser = parse_cookies_from_browser(cookies_from_browser)
            cookie_jar = load_cookies(cookie_file, cookies_from_browser, ydl=None)
            session.cookies = cookie_jar  # type: ignore

    return session


def stat(path):
    try:
        r = requests_session().head(path)
        info = {}

        if 200 <= r.status_code < 400:
            if "content-length" in r.headers:
                info["size"] = int(r.headers["content-length"])

            if "last-modified" in r.headers:
                last_modified = r.headers["last-modified"]
                info["time_modified"] = int(
                    datetime.datetime.strptime(last_modified, "%a, %d %b %Y %H:%M:%S GMT").timestamp()
                )

        elif r.status_code == 404:
            info["time_deleted"] = consts.now()
        else:
            r.raise_for_status()

        return info
    except requests.RequestException as e:
        log.exception("%s could not get metadata", path)
        return {}


def get(args, url, skip_404=True, ignore_errors=False, ignore_429=False, **kwargs):
    s = requests_session(args)
    try:
        response = s.get(url, **kwargs)
    except (
        requests.exceptions.ConnectionError,
        requests.exceptions.Timeout,
        requests.exceptions.ChunkedEncodingError,
        requests.exceptions.ContentDecodingError,
        requests.exceptions.RequestException,
    ):
        raise
    else:
        code = response.status_code

        if 200 <= code < 400:
            return response
        elif code == 404:
            if skip_404:
                log.warning("HTTP404 Not Found: %s", url)
                return None
            else:
                raise FileNotFoundError
        elif ignore_errors and (400 <= code < 429 or 431 <= code < 500):
            return response
        elif ignore_429 and (400 <= code < 500):
            return response
        else:
            response.raise_for_status()

    log.info("Something weird is happening probably: %s", url)
    return response


class PartialContent:
    def __init__(self, url, max_size=1048576):
        self.url = url
        self.max_size = max_size
        self.temp_file = None

    def __enter__(self):
        response = requests_session().get(self.url, stream=True)

        code = response.status_code
        if code == 404:
            log.warning("HTTP404 Not Found: %s", self.url)
            return None
        else:
            response.raise_for_status()

        self.temp_file = tempfile.NamedTemporaryFile(delete=False)

        for chunk in response.iter_content(chunk_size=65536):
            if self.temp_file.tell() < self.max_size:
                self.temp_file.write(chunk)
            else:
                break

        self.temp_file.close()
        return self.temp_file.name

    def __exit__(self, exc_type, exc_value, traceback):
        if self.temp_file:
            os.remove(self.temp_file.name)


def download_embeds(args, soup):
    for img in soup.find_all("img"):
        local_path = Path.cwd() / "images"
        local_path.mkdir(exist_ok=True)
        local_path = local_path / Path(urllib.parse.unquote(img["src"])).name

        response = get(args, img["src"])
        if response:
            with open(local_path, "wb") as f:
                f.write(response.content)

            img["src"] = local_path.relative_to(Path.cwd())  # Update image source to point to local file


def find_date(soup):
    import dateutil.parser

    for text in soup.find_all(text=True):
        try:
            date = dateutil.parser.parse(text)
            return date
        except ValueError:
            pass
    return None


def set_timestamp(headers, path):
    if "Last-Modified" in headers:
        modified_time = datetime.datetime.strptime(headers["Last-Modified"], "%a, %d %b %Y %H:%M:%S GMT")
        mtime = time.mktime(modified_time.timetuple())
        atime = time.time()
        os.utime(path, (atime, mtime))


def load_selenium(args, wire=False):
    if wire:
        import logging

        from seleniumwire import webdriver

        log_levels = [logging.ERROR, logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG]
        for logger_name, logger in logging.root.manager.loggerDict.items():
            if logger_name.startswith("selenium"):
                logging.getLogger(logger_name).setLevel(clamp_index(log_levels, args.verbose - 1))

    else:
        from selenium import webdriver

    xvfb = None  # three states
    if args.verbose < consts.LOG_DEBUG and not getattr(args, "manual", False):
        xvfb = False
        try:
            from pyvirtualdisplay.display import Display

            args.driver_display = Display(visible=False, size=(1280, 720))
            args.driver_display.start()
            xvfb = True
        except Exception:
            pass

    if (which("firefox") or which("firefox.exe") or getattr(args, "firefox", False)) and not getattr(
        args, "chrome", False
    ):
        from selenium.webdriver.firefox.options import Options
        from selenium.webdriver.firefox.service import Service

        service = Service(log_path=tempfile.mktemp(".geckodriver.log"))
        options = Options()
        if Path("selenium").exists():
            options.profile = "selenium"

        options.set_preference("media.volume_scale", "0.0")
        if xvfb is False:
            options.add_argument("--headless")

        args.driver = webdriver.Firefox(service=service, options=options)

        addons = [Path("~/.local/lib/ublock_origin.xpi").expanduser().resolve()]
        if getattr(args, "auto_pager", False):
            addons.append(Path("~/.local/lib/weautopagerize.xpi").expanduser().resolve())

        for addon_path in addons:
            try:
                args.driver.install_addon(str(addon_path))
            except Exception:
                if args.verbose > 0:
                    log.warning("Could not install firefox addon. Missing file %s", addon_path)
                else:
                    log.exception("Could not install firefox addon. Missing file %s", addon_path)

        if getattr(args, "auto_pager", False):
            time.sleep(60)  # let auto-pager initialize

    else:
        from selenium.webdriver.chrome.options import Options

        options = Options()
        if Path("selenium").exists():
            options.add_argument("user-data-dir=selenium")

        options.add_argument("--mute-audio")
        if xvfb is False:
            options.add_argument("--headless=new")

        addons = [Path("~/.local/lib/ublock_origin.crx").expanduser().resolve()]
        if getattr(args, "auto_pager", False):
            addons.append(Path("~/.local/lib/autopager.crx").expanduser().resolve())
        for addon_path in addons:
            try:
                options.add_extension(str(addon_path))
            except Exception:
                if args.verbose > 0:
                    log.warning("Could not install chrome extension. Missing file %s", addon_path)
                else:
                    log.exception("Could not install chrome extension. Missing file %s", addon_path)

        args.driver = webdriver.Chrome(options=options)


def quit_selenium(args):
    args.driver.quit()
    if consts.LOG_DEBUG > args.verbose and not getattr(args, "manual", False):
        try:
            args.driver_display.stop()
        except Exception:
            pass


def safe_unquote(url):
    # https://en.wikipedia.org/wiki/Internationalized_Resource_Identifier
    # we aren't writing HTML so we can unquote

    try:
        parsed_url = urlparse(url)
    except UnicodeDecodeError:
        return url

    def selective_unquote(component, restricted_chars):
        try:
            unquoted = unquote(component, errors="strict")
        except UnicodeDecodeError:
            return component
        # re-quote restricted chars
        return "".join(quote(char, safe="") if char in restricted_chars else char for char in unquoted)

    def unquote_query_params(query):
        query_pairs = parse_qsl(query, keep_blank_values=True)
        return "&".join(
            selective_unquote(key, "=&#") + "=" + selective_unquote(value, "=&#") for key, value in query_pairs
        )

    unquoted_path = selective_unquote(parsed_url.path, ";?#")
    unquoted_params = selective_unquote(parsed_url.params, "?#")
    unquoted_query = unquote_query_params(parsed_url.query)
    unquoted_fragment = selective_unquote(parsed_url.fragment, "")

    new_url = urlunparse(
        (parsed_url.scheme, parsed_url.netloc, unquoted_path, unquoted_params, unquoted_query, unquoted_fragment)
    )

    return new_url


def url_decode(href):
    href = safe_unquote(href)
    up = urlparse(href)
    if up.netloc:
        try:
            href = href.replace(up.netloc, puny_decode(up.netloc), 1)
        except Exception:
            pass
    return href


def safe_quote(url):
    try:
        parsed_url = urlparse(url)
    except UnicodeDecodeError:
        return url

    def selective_quote(component, restricted_chars):
        try:
            quoted = quote(component, errors="strict")
        except UnicodeDecodeError:
            return component
        return "".join(quote(char, safe="%") if char in restricted_chars else char for char in quoted)

    def quote_query_params(query):
        query_pairs = parse_qsl(query, keep_blank_values=True)
        return "&".join(selective_quote(key, "=&#") + "=" + selective_quote(value, "=&#") for key, value in query_pairs)

    quoted_path = selective_quote(parsed_url.path, ";?#")
    quoted_params = selective_quote(parsed_url.params, "?#")
    quoted_query = quote_query_params(parsed_url.query)
    quoted_fragment = selective_quote(parsed_url.fragment, "")

    new_url = urlunparse(
        (parsed_url.scheme, parsed_url.netloc, quoted_path, quoted_params, quoted_query, quoted_fragment)
    )

    return new_url


def url_encode(href):
    href = safe_quote(href)
    up = urlparse(href)
    if up.netloc:
        try:
            href = href.replace(up.netloc, puny_encode(up.netloc).decode(), 1)
        except Exception:
            pass
    return href


def path_tuple_from_url(url):
    url = url_decode(url)
    parsed_url = urlparse(url)
    relative_path = os.path.join(parsed_url.netloc, parsed_url.path.lstrip("/"))
    base_path = os.path.dirname(relative_path)
    filename = os.path.basename(parsed_url.path)
    return base_path, filename


def filename_from_content_disposition(response):
    content_disposition = response.headers.get("Content-Disposition", "")
    if "filename=" in content_disposition:
        msg = Message()
        msg["content-disposition"] = content_disposition
        filename = msg.get_filename()
        if filename:
            return filename
    return None


def url_to_local_path(url, response=None, output_path=None, output_prefix=None):
    base_path, filename = path_tuple_from_url(url)

    if response:
        filename_from_site = filename_from_content_disposition(response)
        if filename_from_site:
            filename = filename_from_site

    if not output_path:
        output_path = filename
        if base_path:
            output_path = os.path.join(base_path, filename)

    output_path = path_utils.clean_path(output_path.encode())

    if output_prefix:
        output_path = os.path.join(output_prefix, output_path)

    return output_path


def download_url(url, output_path=None, output_prefix=None, chunk_size=8 * 1024 * 1024, retry_num=0, max_retries=10):
    if retry_num > max_retries:
        raise RuntimeError(f"Max retries exceeded for {url}")

    session = requests_session()
    r = session.get(url, stream=True)

    if not 200 <= r.status_code < 400:
        log.error(f"Error {r.status_code} {url}")

    remote_size = nums.safe_int(r.headers.get("Content-Length"))

    output_path = url_to_local_path(url, response=r, output_path=output_path, output_prefix=output_prefix)
    if output_path == ".":
        log.warning("Skipping directory %s", url)
        return
    else:
        log.info("Saving %s to %s", url, output_path)

    p = Path(output_path)
    p.parent.mkdir(parents=True, exist_ok=True)
    if p.exists():
        if remote_size:
            local_size = p.stat().st_size
            if local_size == remote_size:
                log.warning(f"Download skipped. File with same size already exists: {output_path}")
                return
            elif local_size < 5242880:  # TODO: check if first few kilobytes match what already exists locally...
                p.unlink()
            else:
                headers = {"Range": f"bytes={local_size}-"}
                r = session.get(url, headers=headers, stream=True)
                if r.status_code != 206:  # HTTP Partial Content
                    p.unlink()
                    r = session.get(url, stream=True)
        else:
            p.unlink()

    try:
        with open(output_path, "ab") as f:
            for chunk in r.iter_content(chunk_size=chunk_size):
                if chunk:
                    f.write(chunk)

        if remote_size:
            downloaded_size = os.path.getsize(output_path)
            if downloaded_size < remote_size:
                msg = f"Incomplete download ({strings.safe_percent(downloaded_size/remote_size)}) {output_path}"
                raise RuntimeError(msg)
    except Exception as e:
        if isinstance(e, OSError):
            if e.errno in consts.EnvironmentErrors:
                raise
        retry_num += 1
        log.info("Retry #%s %s", retry_num, url)
        time.sleep(retry_num)
        return download_url(url, output_path, output_prefix, chunk_size, retry_num)

    set_timestamp(r.headers, output_path)
    return output_path


def get_elements_forward(start, end):
    elements = []
    current = start.next_sibling
    while current and current != end:
        elements.append(current)
        current = current.next_sibling
    return elements


def extract_nearby_text(a_element):
    prev_a = a_element.find_previous("a")
    next_a = a_element.find_next("a")

    before = ""
    if prev_a:
        before = " ".join(s.get_text(strip=True) for s in get_elements_forward(prev_a, a_element))

    after = ""
    if next_a:
        after = " ".join(s.get_text(strip=True) for s in get_elements_forward(a_element, next_a))

    return before, after


def save_html_table(args, html_file):
    import pandas as pd

    dfs = pd.read_html(html_file, extract_links="body", flavor="bs4")
    tables = []
    for df in dfs:
        df = pd_utils.columns_snake_case(df)

        # extract URLs into their own columns
        for col in df.columns:
            if df[col].dtype == "object":
                df[[col, f"{col}_url"]] = pd.DataFrame(df[col].tolist(), index=df.index)
        df.columns = df.columns.astype(str)
        df = df.dropna(axis=1, how="all")  # drop empty columns

        df = pd_utils.convert_dtypes(df)

        tables.append({"table_name": None, "data": df.to_dict(orient="records")})

    tables = db_utils.add_missing_table_names(args, tables)
    for d in tables:
        args.db[d["table_name"]].insert_all(iterables.list_dict_filter_bool(d["data"]), alter=True)


def re_trigger_input(driver):
    from selenium.common.exceptions import NoSuchElementException
    from selenium.webdriver.common.by import By
    from selenium.webdriver.common.keys import Keys

    input_field = None
    for by, name in [
        (By.NAME, "q"),
        (By.NAME, "query"),
        (By.CSS_SELECTOR, "input[type='search']"),
        (By.NAME, "search"),
        (By.NAME, "search-input"),
        (By.NAME, "search-query"),
        (By.NAME, "search-box"),
        (By.ID, "q"),
        (By.ID, "query"),
        (By.ID, "search"),
        (By.ID, "search-input"),
        (By.ID, "search-query"),
        (By.ID, "search-input"),
        (By.ID, "search-box"),
        (By.CLASS_NAME, "q"),
        (By.CLASS_NAME, "query"),
        (By.CLASS_NAME, "search"),
        (By.CLASS_NAME, "search-input"),
        (By.CLASS_NAME, "search-query"),
        (By.CLASS_NAME, "search-input"),
        (By.CLASS_NAME, "search-box"),
        (By.TAG_NAME, "input"),
    ]:
        try:
            input_field = driver.find_element(by, name)
            break
        except NoSuchElementException:
            pass

    if input_field is None:
        return
    else:
        input_field.send_keys(Keys.RETURN)
        driver.implicitly_wait(8)


def selenium_get_page(args, url):
    args.driver.get(url)
    args.driver.implicitly_wait(5)

    if getattr(args, "poke", False):
        re_trigger_input(args.driver)


def scroll_down(driver):
    driver.execute_script("window.scrollTo(0, document.body.scrollHeight);")
    time.sleep(2)
    new_height = driver.execute_script("return document.body.scrollHeight")
    return new_height


def extract_html(url) -> str:
    session = requests_session()
    r = session.get(url, timeout=120, headers=headers)
    r.raise_for_status()
    markup = r.text
    return markup


def selenium_extract_html(driver) -> str:
    # trigger rollover events
    driver.execute_script(
        "(function(){function k(x) { if (x.onmouseover) { x.onmouseover(); x.backupmouseover = x.onmouseover; x.backupmouseout = x.onmouseout; x.onmouseover = null; x.onmouseout = null; } else if (x.backupmouseover) { x.onmouseover = x.backupmouseover; x.onmouseout = x.backupmouseout; x.onmouseover(); x.onmouseout(); } } var i,x; for(i=0; x=document.links[i]; ++i) k(x); for (i=0; x=document.images[i]; ++i) k(x); })()"
    )

    # include Shadow DOM
    html_text = driver.execute_script(
        'function s(n=document.body){if(!n)return"";if(n.nodeType===Node.TEXT_NODE)return n.textContent.trim();if(n.nodeType!==Node.ELEMENT_NODE)return"";let t="";let r=n.cloneNode();n=n.shadowRoot||n;if(n.children.length)for(let o of n.childNodes)if(o.assignedNodes){if(o.assignedNodes()[0])t+=s(o.assignedNodes()[0]);else t+=o.innerHTML}else t+=s(o);else t=n.innerHTML;return r.innerHTML=t,r.outerHTML}; return s()'
    )

    return html_text


def infinite_scroll(driver):
    last_height = driver.execute_script("return document.body.scrollHeight")
    while True:
        new_height = scroll_down(driver)
        yield selenium_extract_html(driver)

        if new_height == last_height:  # last page
            time.sleep(5)  # try once more in case slow page
            new_height = scroll_down(driver)
            if new_height == last_height:
                break
        last_height = new_height

    yield selenium_extract_html(driver)


def construct_search(engine, s):
    s = urllib.parse.quote(s, safe="")
    return engine.replace("%", s, 1)


def construct_absolute_url(base_url, href):
    up = urlparse(href)
    if up.scheme and up.scheme not in ("https", "http", "ftp"):
        return href

    if not up.netloc:
        href = urljoin(base_url + "/", href)

    return href


def is_index(url):
    if url.endswith("/"):
        return True

    patterns = [
        r"/index\.php\?dir=",
        r"/index\.php$",
        r"/index\.html?$",
    ]
    for pattern in patterns:
        if re.search(pattern, url, re.IGNORECASE):
            return True

    return False


def is_subpath(parent_url, child_url):
    child = urlparse(child_url)
    parent = urlparse(parent_url)

    if child.scheme != parent.scheme or child.netloc != parent.netloc:
        return False

    return child_url.startswith(parent_url)


def remove_apache_sorting_params(url):
    parsed_url = urlparse(url)
    query_params = parse_qs(parsed_url.query)

    apache_sorting_keys = ["C", "O"]
    for key in apache_sorting_keys:
        query_params.pop(key, None)
    new_query_string = urlencode(query_params, doseq=True)

    new_url = urlunparse(
        (
            parsed_url.scheme,
            parsed_url.netloc,
            parsed_url.path,
            parsed_url.params,
            new_query_string,
            parsed_url.fragment,
        )
    )

    return new_url


def is_html(url, max_size=15 * 1024 * 1024):
    r = requests_session().get(url, stream=True)

    content_length = r.headers.get("Content-Length")
    if content_length and int(content_length) > max_size:
        return False

    content_type = r.headers.get("Content-Type")
    if content_type and not any(
        s in content_type for s in ("text/html", "text/xhtml", "text/xml", "application/xml", "application/xhtml+xml")
    ):
        return False

    try:
        chunk = next(r.iter_content(max_size + 1))  # one more byte than max_size
        if len(chunk) > max_size:
            return False
    except (requests.RequestException, StopIteration):
        return False

    return True  # if ambiguous, return True


def fake_title(url):
    p = urllib.parse.urlparse(url)
    title = f"{p.netloc} {p.path} {p.params} {p.query}: {p.fragment}"

    if title.startswith("www."):
        title = title[4:]

    title = title.replace("/", " ")
    title = title.replace("?", " ")
    title = title.replace("#", ": ")

    return title.strip()


def get_title(args, url):
    import requests.exceptions
    from bs4 import BeautifulSoup

    try:
        if getattr(args, "local_html", False):
            with open(url) as f:
                html_text = f.read()
            url = "file://" + url
        elif args.selenium:
            web.selenium_get_page(args, url)
            html_text = args.driver.page_source
        else:
            html_text = web.requests_session(args).get(url).text

        soup = BeautifulSoup(html_text, "lxml")
        title = soup.title.text.strip() if soup.title else url
    except requests.exceptions.RequestException as e:
        title = fake_title(url)
    return title



from xklb import usage
from xklb.utils import arggroups, argparse_utils, db_utils


def optimize_db() -> None:
    parser = argparse_utils.ArgumentParser(prog="library optimize", usage=usage.optimize)
    parser.add_argument("--fts", action="store_true")
    parser.add_argument("--force", "-f", action="store_true")
    arggroups.debug(parser)

    arggroups.database(parser)
    args = parser.parse_args()
    arggroups.args_post(args, parser)

    db_utils.optimize(args)



from xklb import usage
from xklb.utils import arggroups, argparse_utils, nums
from xklb.utils.log_utils import log


def parse_utils():
    parser = argparse_utils.ArgumentParser(description="Add arbitrary rows to a SQLITE db", usage=usage.row_add)
    parser.add_argument("--table-name", "--table", "-t", default="media")
    arggroups.debug(parser)

    arggroups.database(parser)
    args, unknown_args = parser.parse_known_args()
    arggroups.args_post(args, parser, create_db=True)
    return args, unknown_args


def parse_unknown_args_to_dict(unknown_args):
    kwargs = {}
    key = None
    values = []

    def get_val():
        if len(values) == 1:
            return nums.safe_int_float_str(values[0])
        else:
            return " ".join(values)

    for arg in unknown_args:
        if arg.startswith("--") or arg.startswith("-"):
            if key is not None:
                kwargs[key] = get_val()  # previous values
                values.clear()
            # Process the new key
            key = arg.strip("-").replace("-", "_")
        else:
            values.append(arg)

    if len(values) > 0:
        kwargs[key] = get_val()

    return kwargs


def row_add():
    args, unknown_args = parse_utils()

    kwargs = parse_unknown_args_to_dict(unknown_args)
    if not kwargs:
        log.error("No data given via arguments")
        raise SystemExit(2)
    args.db[args.table_name].insert(kwargs, alter=True)
